# -*- coding: iso-8859-1 -*-
# circuit.py
# Describes the circuit
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""
Introduction
\"\"\"\"\"\"\"\"\"\"\"\"

A circuit is described in the ahkab simulator by an instance of the
:class:`Circuit` class.

This class holds everything is needed to simulate the circuit, except
the specification of the analyses to be performed.

To rewrite a netlist from a Circuit instance see the
:mod:`ahkab.printing` module.

The Circuit
\"\"\"\"\"\"\"\"\"\"\"

A circuit is derived from a list which contains all its elements.

Conceptually, every time an element is to be inserted in the circuit,
two operations have to be performed:

* The element must be appended to the ``Circuit`` instance.
* Its connections should be ensure checking that
  the nodes the element refers to are indeed existing circuit nodes.

To simplify the operation of adding a component to a ``Circuit``,
the following convenience methods are provided to the user to add and
remove most elements to the circuit:

* :func:`Circuit.add_resistor`
* :func:`Circuit.add_capacitor`
* :func:`Circuit.add_inductor`
* :func:`Circuit.add_vsource`
* :func:`Circuit.add_isource`
* :func:`Circuit.add_diode`
* :func:`Circuit.add_mos`
* :func:`Circuit.add_cccs`
* :func:`Circuit.add_vcvs`
* :func:`Circuit.add_vccs`
* :func:`Circuit.add_user_defined`
* :func:`Circuit.remove_elem`

Example:

.. code-block:: python

    mycircuit = circuit.Circuit(title="Example circuit", filename=None)
    # no filename since there will be no deck associated with this circuit.
    # get the ref node (gnd)
    gnd = mycircuit.get_ground_node()
    # add a node named n1 and a 600 ohm resistor connected between n1 and gnd
    mycircuit.add_resistor(part_id="R1", n1="n1", n2=gnd, R=600)

Refer to the methods help for additional information.

Nodes
\"\"\"\"\"

The nodes are internally stored in the following way: we assign to each node an
internal ID, independetly from its external identifier used in the netlist.
Those IDs are integers.

The simulator uses always the internal names. When the results are
presented to the user, the internal node is not showed, the external
identifier (or external node name) is printed instead.

This is done through:

.. code-block:: python

    my_circuit = Circuit()
    ...
    [ init code ]
    ...
    print "This is a node" + my_circuit.nodes_dict[int_node]

.. rubric:: Internal only nodes

The number of internal only nodes (added automatically by the simulator)
is held in ``Circuit.internal_nodes``. That value shouldn't be changed by
hand.

Device models
\"\"\"\"\"\"\"\"\"\"\"\"\"

Non-linear elements have their operation described by specialized routines
held in their module.

They are stored in ``Circuit.models`` (of type dict), the following methods
are provided to add and remove device models to a Circuit instance.

* :func:`Circuit.add_model`
* :func:`Circuit.remove_model`

Reference
\"\"\"\"\"\"\"\"\"\"\"\"\"\"\"\"\"\"\""\"\"\"\"\"\"\"\"\"\"\"\"\"\"\"\"\"\"\"

"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import copy
import math

from . import devices
from . import diode
from . import ekv
from . import mosq
from . import printing
from . import py3compat
from . import switch

# will be added here by netlist_parser and circuit instances
user_defined_modules_dict = {}

class Circuit(list):
    """The circuit class.

    **Parameters:**

    title : string
        The circuit title.

    filename : string, optional

        .. deprecated:: 0.09

        If the circuit instance corresponds to a netlist file on disk,
        set this to the netlist filename.

    """
    def __init__(self, title, filename=None):
        self.title = title
        self.filename = filename
        self.nodes_dict = {}  # {int_node:ext_node, int_node:ext_node}
        self.internal_nodes = 0
        self.models = {}
        self.gnd = '0'

    def __str__(self):
        s = "* " + self.title + "\n"
        for elem in self:
            s += elem.get_netlist_elem_line(self.nodes_dict) + "\n"
        return s[:-1]

    def create_node(self, name):
        """Creates a new circuit node

        If there is a node in the circuit with the same name, ValueError is
        raised.

        **Parameters:**

        name : string
            the _unique_ identifier of the node.

        **Returns:**

        node : string
            the _unique_ identifier of the node, to be used for subsequent
            element declarations, for example.

        :raises ValueError: if a new node with the given id cannot be created,
          for example because a node with the same name already exists in the
          circuit. The only exception is the ground node, which has the
          reserved id ``'0'``, and for which this method won't raise any
          exception.
        :raises TypeError: if the parameter ``name`` is not of "text" type (what
          that means exactly depends on which version of Python you are using.)

        """
        if type(name) not in py3compat.string_types:
            raise TypeError("The node %s should have been of text type" %
                            name)
        got_ref = 0 in self.nodes_dict
        if name not in self.nodes_dict:
            if name == '0':
                int_node = 0
            else:
                int_node = int(len(self.nodes_dict)/2) + 1*(not got_ref)
            self.nodes_dict.update({int_node:name})
            self.nodes_dict.update({name:int_node})
        else:
            raise ValueError('Impossible to create new node %s: node exists!'
                             % name)
        return name

    def add_node(self, ext_name):
        """Adds the supplied node to the circuit, if needed.

        When a 'normal' (not the reference) node is added, a internal
        name (or label) is assigned to it.

        The nodes labels are stored in ``Circuit.nodes_dict``, as a dictionary of pairs
        like ``{int_node:ext_node}``.

        Those internal names are integers, by definition, and are generated
        starting from 1, then 2, 3, 4, 5...
        The integer ``0`` is reserved for the reference node (gnd), which is required
        for the circuit to be non-pathological and has ``ext_name=str(int_name)='0'``.

        Notice that this method doesn't halt or print errors if the node is already been
        added previsiously. It simply returns the internal node name assigned to it.

        **Parameters:**

        ext_name : string
            The unique identifier of the node.

        **Returns:**

        int_name : string
            the *unique* *internal* ciecuit identifier of the node.

        :raises TypeError: if the parameter ``ext_name`` is not of "text" type
          (what that means exactly depends on which version of Python you are
          using.)
        """
        # must be text (str unicode...)
        if type(ext_name) not in py3compat.string_types:
            raise TypeError("The node %s should have been of text type" %
                            ext_name)
        # test: do we already have it in the dictionary?
        if ext_name not in self.nodes_dict:
            if ext_name == '0':
                int_node = 0
            else:
                got_ref = 0 in self.nodes_dict
                int_node = int(len(self.nodes_dict)/2) + 1*(not got_ref)
            self.nodes_dict.update({int_node:ext_name})
            self.nodes_dict.update({ext_name:int_node})
        else:
            int_node = self.nodes_dict[ext_name]
        return int_node

    def new_internal_node(self):
        """Generate implicit internal nodes.

        Some devices are made of a group of other devices, connected by
        "internal only" nodes, which have the prefix ``'INT'`` and the
        simulator treats specially, hiding them from the user if not
        explicitly asked about them.

        This method generates the external names for such nodes and inserts them
        in the circuit.

        **Returns:**

        ext_node : string
            The corresponding external node id.
        """

        ext_node = "INT" + str(self.internal_nodes)
        self.internal_nodes = self.internal_nodes + 1
        self.create_node(ext_node)
        return ext_node

    def get_nodes_number(self):
        """Returns the number of nodes in the circuit"""
        return int(len(self.nodes_dict)/2)

    def is_int_node_internal_only(self, int_node):
        """Check whether an internal node is an "internal only node" or not.

        **Parameters:**

        int_node : int
            The internal only node to be checked.

        **Returns:**

        chk : boolean
            The result of the check.

        :raises TypeError: if the supplied node is not an ``int``. Typically
          this happens when the method is called with an *external* name.
        """
        if type(int_node) is not int:
            raise TypeError('Expecting an INTERNAL node of type int, got %s.' %
                            type(int_node))
        return self.nodes_dict[int_node].find("INT") > -1

    def is_nonlinear(self):
        """Check whether the circuit is non-linear or not.

        **Returns:**

        chk : boolean
            The result of the check.
        """
        for elem in self:
            if elem.is_nonlinear:
                return True
        return False

    def get_locked_nodes(self):
        """Get all nodes connected to non-linear elements.

        This list is meant to be passed to ``dc_solve`` or ``mdn_solver`` to be
        used in ``get_td`` to evaluate the damping coefficient in a
        Newton-Rhapson iteration.

        **Returns:**

        locked_nodes : list
            A list of internal nodes.
        """
        locked_nodes = []
        nl_elements = [elem for elem in self if elem.is_nonlinear]
        for elem in nl_elements:
            oports = elem.get_output_ports()
            for index in range(len(oports)):
                ports = elem.get_drive_ports(index)
                for port in ports:
                    locked_nodes.append(port)
        return locked_nodes

    def ext_node_to_int(self, ext_node):
        """This function returns the integer id associated with an external node id.

        **Parameters:**

        ext_node : string
            The external node id to be converted.

        **Returns:**

        int_node : int
            The internal node associated.

        """
        return self.nodes_dict[ext_node]

    def int_node_to_ext(self, int_node):
        """This function returns the string id associated with the integer internal node id
        ``int_node``.

        **Parameters:**

        int_node : int
            The internal node id to be converted.

        **Returns:**

        ext_node : string
            the string id associated with ``int_node``.
        """
        return self.nodes_dict[int_node]

    def has_duplicate_elem(self):
        """Self-check for duplicate elements.

        No circuit should ever have duplicate elements
        (ie elements with the same ``part_id``).

        **Returns:**

        chk : boolean
            The result of the check.
        """
        all_ids = tuple(map(lambda e: e.part_id, self))
        return len(set(all_ids)) != len(all_ids)

    def get_ground_node(self):
        """Returns the reference node, AKA GND."""
        return '0'

    def get_elem_by_name(self, part_id):
        """Get a circuit element from its ``part_id`` value.

        If no matching element is found, the method returns
        ``None``. This may change in the future.

        **Parameters:**

        part_id : string
            The ``part_id`` of the element

        **Returns:**

        elem : circuit element
            Depending whether a matching element was found or not.

        :raises ValueError: if the element is not found.
        """
        for e in self:
            if e.part_id.lower() == part_id.lower():
                return e
        raise ValueError('Element %s not found' % part_id)

    def add_model(self, model_type, model_label, model_parameters):
        """Add a model to the available circuit models.

        **Parameters:**

        model_type : string
            the model type (eg "ekv"). Right now, the possible values are:
            ``"mosq"``, ``"ekv"``, ``"diode"``, ``"sw"``.

        model_label : string
            a unique identifier for the model being added (eg. ``"nch1"``).

        model_parameters: dict
            a dictionary holding the parameters to be supplied to the
            model to instantiate it.

        """

        if 'name' not in model_parameters:
            model_parameters.update({'name':model_label})
        if model_type == "ekv":
            model_iter = ekv.ekv_mos_model(**model_parameters)
            model_iter.name = model_label
        elif model_type == "mosq":
            model_iter = mosq.mosq_mos_model(**model_parameters)
            model_iter.name = model_label
        elif model_type == "diode":
            model_iter = diode.diode_model(**model_parameters)
            model_iter.name = model_label
        elif model_type == "sw":
            model_iter = switch.vswitch_model(**model_parameters)
            model_iter.name = model_label
        else:
            raise CircuitError("Unknown model type %s" % (model_type,))
        self.models.update({model_label: model_iter})

    def remove_model(self, model_label):
        """Remove a model from the available models.

        **Parameters:**

        model_label : string
            the unique identifier corresponding to the model
            being removed.

        .. note::

            This method currently silently ignores models that are not defined.

        """
        if self.models is not None and model_label in self.models:
            del self.models[model_label]
        # should print a warning here

    def add_resistor(self, part_id, n1, n2, value):
        """Adds a resistor to the circuit.

        The resistor instance is added to the circuit elements
        and connected to the provided nodes. If the nodes are not
        found in the circuit, they are created and added as well.

        **Parameters:**

        part_id : string
            the resistor part_id (eg "R1"). The first letter is replaced by an R

        n1, n2 : string
            the nodes to which the resistor is connected.

        value : float,
            The resistance between ``n1`` and ``n2`` in Ohm.

        .. seealso::

            :func:`add_resistor`, :func:`add_capacitor`,
            :func:`add_inductor`, :func:`add_vsource`, :func:`add_isource`,
            :func:`add_diode`, :func:`add_mos`, :func:`add_vcvs`, :func:`add_vccs`,
            :func:`add_cccs`, :func:`add_user_defined`, :func:`remove_elem`

        """
        n1 = self.add_node(n1)
        n2 = self.add_node(n2)

        if value == 0:
            raise CircuitError("ZERO-valued resistors are not allowed.")

        elem = devices.Resistor(part_id=part_id, n1=n1, n2=n2, value=value)
        self.append(elem)

    def add_capacitor(self, part_id, n1, n2, value, ic=None):
        """Adds a capacitor to the circuit.

        The capacitor instance is added to the circuit elements
        and connected to the provided nodes. If the nodes are not
        found in the circuit, they are created and added as well.

        **Parameters:**

        part_id : string
            The capacitor part_id (eg "C1"). The first letter is always C.

        n1, n2 : string
            The nodes to which the element is connected.

        value : float
            The capacitance value.

        ic : float, optional
            The initial condition, if any. See the simulation docs for
            how this affects the results.

        .. seealso::
            :func:`add_resistor`,
            :func:`add_inductor`, :func:`add_vsource`, :func:`add_isource`,
            :func:`add_diode`, :func:`add_mos`, :func:`add_vcvs`, :func:`add_vccs`,
            :func:`add_cccs`, :func:`add_user_defined`, :func:`remove_elem`

        """
        if value == 0:
            raise CircuitError("ZERO-valued capacitors are not allowed.")

        n1 = self.add_node(n1)
        n2 = self.add_node(n2)

        elem = devices.Capacitor(part_id=part_id, n1=n1, n2=n2, value=value, ic=ic)

        self.append(elem)

    def add_inductor(self, part_id, n1, n2, value, ic=None):
        """Adds an inductor to the circuit.

        The inductor instance is added to the circuit elements
        and connected to the provided nodes. If the nodes are not
        found in the circuit, they are created and added as well.

        **Parameters:**

        part_id : string
            The inductor part_id (eg "Lfilter"). The first letter is always L.

        n1, n2 : string
            The nodes to which the element is connected. Eg. ``"in"`` or ``"out_a"``.

        value : float
            The inductance value.

        ic : float, optional
            Initial condition, see simulation types for how this affects
            the results.

        .. seealso::

            :func:`add_resistor`, :func:`add_capacitor`,
            :func:`add_inductor`, :func:`add_vsource`, :func:`add_isource`,
            :func:`add_diode`, :func:`add_mos`, :func:`add_vcvs`, :func:`add_vccs`,
            :func:`add_cccs`, :func:`add_user_defined`, :func:`remove_elem`
        """

        n1 = self.add_node(n1)
        n2 = self.add_node(n2)

        elem = devices.Inductor(part_id=part_id, n1=n1, n2=n2, value=value, ic=ic)

        self.append(elem)

    def add_inductor_coupling(self, part_id, L1, L2, value):
        """Add a coupling between two inductors.

        **Parameters:**

        part_id : string
            The part ID for the inductor coupling device. Eg. ``'K1'``,
            the first letter is always ``'K'``.
        L1 : string
            The part ID of the first inductor to be coupled.
        L2 : string
            The part ID of the second inductor to be coupled.
        value : float
            The ``k`` value of the mutual coupling coefficient.
            Its value must be greater than zero and lesser or equal to``1``
            or instability ensues.
        """
        L1elem, L2elem = None, None

        for e in self:
            if isinstance(e, devices.Inductor) and (L1 == e.part_id):
                L1elem = e
            elif isinstance(e, devices.Inductor) and (L2 == e.part_id):
                L2elem = e

        if L1elem is None or L2elem is None:
            error_msg = "One or more coupled inductors for %s were not found: %s (found: %s), %s (found: %s)." % \
                (part_id, L1, L1elem is not None, L2, L2elem is not None)
            raise ValueError(error_msg)

        M = math.sqrt(L1elem.value * L2elem.value) * value

        elem = devices.InductorCoupling(part_id=part_id, L1=L1, L2=L2, K=value, M=M)
        L1elem.coupling_devices.append(elem)
        L2elem.coupling_devices.append(elem)

        self.append(elem)

    def add_vsource(self, part_id, n1, n2, dc_value, ac_value=0, function=None):
        """Adds a voltage source to the circuit (also takes care that the nodes
        are added as well).

        **Parameters:**

        part_id : string
            The voltage source part_id (eg "VA"). The first letter is always V.
        n1, n2 : string
            The nodes to which the element is connected. Eg. ``"in"`` or
            ``"out_a"``.
        dc_value : float
            DC voltage value
        ac_value : float, optional
            AC voltage value, defaults to 0.
        function : function, optional
            Time function. See devices.py for built-in options.
        """
        n1 = self.add_node(n1)
        n2 = self.add_node(n2)

        elem = devices.VSource(part_id=part_id, n1=n1, n2=n2, dc_value=dc_value,
                               ac_value=ac_value)

        if function is not None:
            elem.is_timedependent = True
            elem._time_function = function

        self.append(elem)

    def add_isource(self, part_id, n1, n2, dc_value, ac_value=0, function=None):
        """Adds a current source to the circuit (also takes care that the nodes
        are added as well).

        **Parameters:**

        part_id : string
            The current source ID (eg ``"IA"`` or ``"I3"``). The first letter
            is always I.
        n1, n2 : string
            The nodes to which the element is connected, eg. ``"in"`` or ``"out1"``.
        dc_value : float
            DC current value.
        ac_value :float, optional
            AC current value, defaults to 0.
        function : function, optional
            Time function. See devices.py for built-in options.
        """
        n1 = self.add_node(n1)
        n2 = self.add_node(n2)

        elem = devices.ISource(part_id=part_id, n1=n1, n2=n2, dc_value=dc_value,
                               ac_value=ac_value)

        if function is not None:
            elem.is_timedependent = True
            elem._time_function = function

        self.append(elem)

    def add_diode(self, part_id, n1, n2, model_label, models=None, Area=None,
                  T=None, ic=None, off=False):
        """Adds a diode to the circuit (also takes care that the nodes
        are added as well).

        **Parameters:**

        part_id : string
            The diode ID (eg "D1"). The first letter is always D.
        n1, n2 : string
            the nodes to which the element is connected. eg. ``"in"`` or
            ``"out_a"``
        model_label : string
            The diode model identifier. The model needs to be added
            first, then the elements using it.
        models : dict, optional
            List of available model instances. If not set or ``None``,
            the circuit models will be used (recommended).
        Area : float, optional
            Scaled device area (optional, defaults to 1.0)
        T : float, optional
            Operating temperature (no temperature dependence yet)
        ic : float, optional
            Initial condition (not really implemented yet)
        off : bool, optional
            Consider the diode to be initially off.
        """
        n1 = self.add_node(n1)
        n2 = self.add_node(n2)
        if models is None:
            models = self.models
        if model_label not in models:
            raise ModelError("Unknown diode model id: " + model_label)

        elem = diode.diode(part_id=part_id, n1=n1, n2=n2, model=models[
                           model_label], AREA=Area, T=T, ic=ic, off=off)
        self.append(elem)


    def add_mos(self, part_id, nd, ng, ns, nb, w, l, model_label, models=None,
                m=1, n=1):
        """Adds a mosfet to the circuit (also takes care that the nodes
        are added as well).

        **Parameters:**

        part_id : string
            The mos part_id (eg "M1"). The first letter is always M.
        nd : string
            The drain node.
        ng : string
            The gate node.
        ns : string
            The source node.
        nb : string
            The bulk node.
        w : float
            The gate width.
        l : float
            The gate length.
        model_label : string
            The model identifier.
        models : dict, optional
            The circuit models.
        m : int, optional
            Shunt multiplier value. Defaults to 1.
        n : int, optional
            Series multiplier value, not always supported. Defaults to 1.
        """
        nd = self.add_node(nd)
        ng = self.add_node(ng)
        ns = self.add_node(ns)
        nb = self.add_node(nb)

        if models is None:
            models = self.models

        if model_label not in models:
            raise ModelError("Unknown model id: " + model_label)

        if isinstance(models[model_label], ekv.ekv_mos_model):
            elem = ekv.ekv_device(part_id, nd, ng, ns, nb, w, l,
                                  models[model_label], m, n)
        elif isinstance(models[model_label], mosq.mosq_mos_model):
            elem = mosq.mosq_device(part_id, nd, ng, ns, nb, w, l,
                                    models[model_label], m, n)
        else:
            raise Exception("Unknown model type for " + model_label)

        self.append(elem)


    def add_cccs(self, part_id, n1, n2, source_id, value):
        """Adds a current-controlled current source (CCCS) to the circuit

        This method takes care that its nodes are added as well.

        **Parameters:**

        part_id : string
            The cccs ID (eg ``'F1'``). The first letter is always ``'F'``.
        n1, n2 : strings
            The output port nodes, where the output current is
            forced. Eg. "outp", "outm" or "out_a", "out_b".
        source_id : string
            The voltage source to be used to sense the current that drives
            the output. Eg. ``'V1'``.
        value : float
            The proportionality factor between input (:math:`I_s`) and output
            (:math:`I_o`) currents. Mathematically:

            .. math::

                I_o = \\alpha I_s

        .. seealso::

            :class:`ahkab.devices.FISource`

        """
        # Add the nodes, this is SAFE: if a node is already known to the circuit,
        # the methods will just ignore the request.
        n1 = self.add_node(n1)
        n2 = self.add_node(n2)
        # instantiate the element
        elem = devices.FISource(part_id=part_id, n1=n1, n2=n2,
                                source_id=source_id, value=value)
        # add it!
        self.append(elem)

    def add_ccvs(self, part_id, n1, n2, source_id, value):
        """Adds a current-controlled voltage source (CCCS) to the circuit

        This method takes care that its nodes are added as well.

        **Parameters:**

        part_id : string
            The cccs ID (eg ``'H1'``). The first letter is always ``'H'``.
        n1, n2 : strings
            The output port nodes, where the output current is
            forced. Eg. "outp", "outm" or "out_a", "out_b".
        source_id : string
            The voltage source to be used to sense the current that drives
            the output voltage. Eg. ``'V1'``.
        value : float
            The proportionality factor between the sense current :math:`I_s`
            flowing into the ``source_id`` voltage source (input) and output voltage.
            Mathematically:

            .. math::

                Vn_1 - Vn_2 = \\alpha I_s

        .. seealso::

            :class:`ahkab.devices.EVSource`,
            :class:`ahkab.devices.FISource`

        """
        # Add the nodes, this is SAFE: if a node is already known to the circuit,
        # the methods will just ignore the request.
        n1 = self.add_node(n1)
        n2 = self.add_node(n2)
        # instantiate the element
        elem = devices.HVSource(part_id=part_id, n1=n1, n2=n2,
                                source_id=source_id, value=value)
        # add it!
        self.append(elem)

    def add_vcvs(self, part_id, n1, n2, sn1, sn2, value):
        """Adds a voltage-controlled voltage source (vcvs) to the circuit

        This method also takes care that its nodes are added as well.

        **Parameters:**

        part_id : string
            The vcvs ID (eg "E1"). The first letter is always E.
        n1, n2 : string
            The output port nodes, where the output voltage is
            forced. Eg. "outp", "outm" or "out_a", "out_b".
        sn1, sn2 : string
            The input port nodes, where the input voltage is
            read. Eg. "inp", "inm" or "in_a", "in_b".
        alpha : float
            The proportionality factor between input and output voltages is
            given by the relationship:

            .. math::

                V(out_p) - V(out_n) = \\alpha \\cdot (V(in_p) - V(in_n))

        """

        n1 = self.add_node(n1)
        n2 = self.add_node(n2)
        sn1 = self.add_node(sn1)
        sn2 = self.add_node(sn2)

        elem = devices.EVSource(part_id=part_id, n1=n1, n2=n2, sn1=sn1, sn2=sn2,
                                value=value)

        self.append(elem)


    def add_vccs(self, part_id, n1, n2, sn1, sn2, value):
        """Adds a voltage-controlled current source (VCCS) to the circuit

        This method also takes care that its nodes are added as well.

        **Parameters:**

        part_id : string
            The VCCS ID (eg ``"G1"``). The first letter is always ``'G'``.
        n1, n2 : string
            The output port nodes, where the output current is
            forced. Eg. "outp", "outm" or "out_a", "out_b".
            The passive convention is used as everywhere else in the simulator:
            a positive current flows into ``n1`` and out of ``n2``.
        sn1, sn2 : string
            The input port nodes, where the input voltage is
            sensed. Eg. "inp", "inm" or "in_a", "in_b".
        value : float
            The proportionality factor between input and output voltages,
            which are related by the equality:

            .. math::

                I_o = alpha * \\left[V(inp) - V(inn)\\right]

        """

        n1 = self.add_node(n1)
        n2 = self.add_node(n2)
        sn1 = self.add_node(sn1)
        sn2 = self.add_node(sn2)

        elem = devices.GISource(part_id=part_id, n1=n1, n2=n2, sn1=sn1, sn2=sn2,
                                value=value)

        self.append(elem)

    def add_switch(self, part_id, n1, n2, sn1, sn2, ic, model_label, models=None):
        """Adds a voltage-controlled or current-controlled switch to the circuit

        This method also takes care that its nodes are added to the circuit as
        well, if necessary.

        **Notice:**

        - Current-controlled switches are not yet implemented. If you try to add
          one, you'll trigger an error. If you got a bit of time to spare,
          patches are welcome.
        - The switches ``part_id`` should begin with ``'S'`` for
          voltage-controlled switches and with ``'W'`` for current-controlled
          switches.
        - The actual behavior is set by the model. Make sure you supply a
          voltage-controlled switch model for a voltage-controlled switch and
          the appropriate type of model for the current-controlled switch.
          Mixing them up will go undetected.

        **Parameters:**

        part_id : string
            the switch ID (eg ``"S1"`` - voltage-controlled - or ``"Wa"`` -
            current-controlled). The first letter is always ``S`` or ``W``.
        n1, n2 : string
            the output port nodes, where the switch is connected. Eg. ``"out1"``,
            ``"out2"`` or ``"n_a"``, ``"n_b"``.
        sn1, sn2 : string
            The input port nodes, where the input voltage is
            read. Eg. "inp", "inm" or "in_a", "in_b".
        ic : boolean
            The initial conditions for transient simulation. Not currently
            implemented!
        model_label : string
            The switch model identifier. The model needs to be added
            first, then the elements using it.
        models : dict, optional
            A dictionary assembled as (identifier:instance), containing all the available model
            instances. If not set or ``None``, the circuit models will be used (recommended).
        """

        n1 = self.add_node(n1)
        n2 = self.add_node(n2)
        sn1 = self.add_node(sn1)
        sn2 = self.add_node(sn2)

        if models is None:
            models = self.models
        if model_label not in models:
            raise ModelError("Unknown switch model id: " + model_label)

        elem = switch.switch_device(part_id=part_id, n1=n1, n2=n2, sn1=sn1,
                                    sn2=sn2, model=models[model_label])

        self.append(elem)

    def add_user_defined(self, module, label, param_dict):
        """Adds a user defined element.

        In order for this to work, you should write a module that supplies the
        elem class.

        XXX WRITE DOC
        """

        if module_name in circuit.user_defined_modules_dict:
            module = circuit.user_defined_modules_dict[module_name]
        else:
            fp, pathname, description = imp.find_module(module_name)
            module = imp.load_module(module_name, fp, pathname, description)
            circuit.user_defined_modules_dict.update({module_name: module})

        elem_class = getattr(module, label)

        param_dict.update({"convert_units": convert_units})
        param_dict.update({"circuit_node": self.add_node})

        elem = elem_class(**param_dict)
        elem.part_id = "y%s" % part_id[1:]

        # call check() if supported
        if hasattr(elem, "check"):
            selfcheck_result, error_msg = elem.check()
            if not selfcheck_result:
                raise NetlistParseError("module: " + module_name + \
                                        " elem type: " + elem_type_name + \
                                        " error: " + error_msg)

        self.append(elem)

    def remove_elem(self, elem_or_id):
        """Removes an element from the circuit and takes care that no
        "orphan" nodes are left.

        .. note::

            Support for removing elements is experimental.

        **Parameters:**

        elem_or_id : string or circuit element
            You may pass as first element, alternatively, either the ``part_id``
            of the element to be removed or the element itself.

        The method will also take care of purging from the circuit nodes that
        are left orphan, ie with no elements connected.

        :raises ValueError: if no such element is found in the circuit.

        """
        if type(elem_or_id) in py3compat.string_types:
            #we got a part_id, we need the element
            elem = self.get_elem_by_name(elem_or_id)
        else:
            # we got the element
            elem = elem_or_id
        self.remove(elem)
        nodes = []
        if hasattr(elem, 'n1') and elem.n1 != 0:
            nodes = nodes + [elem.n1]
        if hasattr(elem, 'n2') and elem.n2 != 0 and elem.n2 not in nodes:
            nodes = nodes + [elem.n2]
        if elem.is_nonlinear:
            for n1, n2 in elem.ports:
                if n1 != 0 and n1 not in nodes:
                    nodes = nodes + [n1]
                if n2 != 0 and n2 not in nodes:
                    nodes = nodes + [n2]
        # check if then nodes are needed by other elements
        remove_list = copy.copy(nodes)
        for n in nodes:
            for e in self:
                if hasattr(elem, 'n1') and e.n1 == n or \
                   hasattr(elem, 'n2') and e.n2 == n or \
                   hasattr(elem, 'sn1') and e.sn1 == n or \
                   hasattr(elem, 'sn2') and e.sn2 == n:
                    remove_list.remove(n)
                    break
                if elem.is_nonlinear:
                    oports = elem.get_output_ports()
                    # check output ports
                    for n1, n2 in oports:
                        if n1 == n or n2 == n:
                            remove_list.remove(n)
                            break
                    if n not in remove_list:
                        break
                    # check the ports that drive them
                    for i in range(len(oports)):
                        dports = elem.get_drive_ports(i)
                        for n1, n2 in oports:
                            if n1 == n or n2 == n:
                                remove_list.remove(n)
                                break
                        if n not in remove_list:
                            break

        for n in remove_list:
            self.nodes_dict.pop(self.nodes_dict[n])
            self.nodes_dict.pop(n)

    def find_vde_index(self, elem_or_id, verbose=3):
        """Finds a voltage-defined element MNA index.

        **Parameters:**

        elem_or_id : string or circuit element
            You may pass as first element, alternatively, either the ``part_id``
            of the element whose index is being requested (eg. 'V1') or the
            element itself.
            Notice the ``part_id`` includes both the id letter (eg. 'V') and the
            description (eg. '1').
        verbose : int
            The verbosity level, from 0 (silent) to 6 (debug).

        **Returns:**

        indx : int
            The index.

        :raises ValueError: if no such element is in the circuit.
        """
        if type(elem_or_id) not in py3compat.string_types:
            # we got an element
            part_id = elem_or_id.part_id
        else:
            # we got a string corresponding to the part_id of an element
            part_id = elem_or_id
        vde_index = 0
        for elem in self:
            if is_elem_voltage_defined(elem):
                if elem.part_id.upper() == part_id.upper():
                    break
                else:
                    vde_index += 1
        else:
            raise ValueError(("find_vde_index(): element %s was not found." +\
                              " This is a bug.") % (part_id,))
        printing.print_info_line(("%s found at index %d" % (part_id,
                                                            vde_index), 6),
                                 verbose)
        return vde_index

    def find_vde(self, index):
        """Finds a voltage-defined element from its MNA KVL index

        **Parameters:**

        index : int
            The element index in the KVL equations.

        **Returns:**

        e : circuit element (an instance of a subclass of Component)
            The element corresponding to ``index``.

        :raises IndexError: if no element corresponds to such an index.
        """
        index = index - len(self.nodes_dict)/2 + 1
        ni = 0
        for e in self:
            if is_elem_voltage_defined(e):
                if index == ni:
                    break
                else:
                    ni = ni + 1
        else: #executed if no break occurred
            raise IndexError('No element corresponds to vde index %d' %
                             (index + len(self.nodes_dict)/2 - 1))
        return e


# STATIC METHODS
def is_elem_voltage_defined(elem):
    """Check if an element needs its own KCL equation

    **Parameters:**

    elem : Component
        The element to be checked.

    **Returns:**

    chk : bool
        ``True`` if ``elem`` is a voltage source, an inductor, a voltage-controlled
        voltage source or a current-controlled voltage source. ``False`` otherwise.
    """
    if isinstance(elem, devices.VSource) or isinstance(elem, devices.EVSource) or \
        isinstance(elem, devices.HVSource) or isinstance(elem, devices.Inductor) \
            or (hasattr(elem, "is_voltage_defined") and elem.is_voltage_defined):
        return True
    else:
        return False


class NodeNotFoundError(Exception):
    """Circuit Node exception."""
    pass


class CircuitError(Exception):
    """General circuit assembly exception."""
    pass


class ModelError(Exception):
    """Model not found exception."""
    pass


class subckt:
    """This class holds the necessary information about a circuit.

    An instance of this class is returned by:

    :func:`ahkab.netlist_parser.parse_sub_declaration`

    **Parameters:**

    name : string
        The subcircuit definition label.
    code : string
        The netlist code that can be instantiated have a circuit
        instance.
    connected_nodes_list : list
        A list of nodes that are used in the circuit and that are
        meant to be connected to the external circuit.

    Notice that in the current implementation, the GND node (0)
    is *always* global.

    """
    def __init__(self, name, code, connected_nodes_list):
        self.name = name
        self.connected_nodes_list = connected_nodes_list
        self.code = code


class _circuit_wrapper:
    """Fictious circuit class, meant to wrap subcircuits.

    Not meant for end users at this stage.

    Rationale:

    Within a subcircuit, the nodes name are fictious.
    All nodes have to be renamed before a subcircuit instance
    may be insterted in a circuit (all our circuits are flat in
    memory for now).

    The nodes of the subcircuit that are connected to the
    nodes of the circuit have to be renamed to them, those
    that are not referenced there need to be renamed in order
    for them ot be unique.

    This class wraps a circuit object and performs the conversion
    _before_ calling ``circ.add_node()``.

    While instatiating/calling a subcircuit wrap the circuit instance
    in this.

    **Parameters:**

    circ : circuit instance
        The main circuit. Remember that all our assembled circuits
        are flat in memory.
    connection_nodes_dict : dictionary
        The dictionary mapping internal nodes to global, circuit-wide nodes.
    subckt_name : string
        The subcircuit instance name. The first letter must always be ``'X'``.
    subckt_label : string
        The label of the subcircuit that is being instantiated.

    """

    def __init__(self, circ, connection_nodes_dict, subckt_name, subckt_label):
        self.circ = circ
        self.prefix = subckt_label + "-" + subckt_name + "-"
        self.subckt_node_filter_dict = {}
        self.subckt_node_filter_dict.update(connection_nodes_dict)
        self.subckt_node_filter_dict.update({'0': '0'})

    def add_node(self, ext_name):
        """We want to perform the following conversions:

        * connected node in the subcircuit -> node in the upper circuit
        * local-only node of the subcircuit -> rename it to something unique
        * REF (0) -> REF (0)

        And then call ``circ.add_node()``.
        """
        if ext_name not in self.subckt_node_filter_dict:
            self.subckt_node_filter_dict.update(
                {ext_name: self.prefix + ext_name})
        int_node = self.circ.add_node(self.subckt_node_filter_dict[ext_name])
        return int_node



#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
# ahkab.py
# The frontend of the simulator
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

""" ahkab is an easy electronic circuit simulator.
"""

from __future__ import print_function

import os

# Package testing can be done remotely, without display. This would make
# matplotlib fail (and, sometimes, the test as well).
# If we check for $DISPLAY, that makes us probably lose in portability,
# because does Windows have the DISPLAY env variable defined?
try:
    import matplotlib
    if os.system('python -c "import matplotlib.pyplot as plt;plt.figure()"'):
        # no display, use a display-less backend
        matplotlib.use('Agg')
except ImportError:
    # plotting.py will complain about this later on.
    pass

from .ahkab import new_op, new_dc, new_tran, new_ac, new_pss, new_pz
from .ahkab import new_symbolic, queue, run, new_x0, icmodified_x0
from .ahkab import get_op_x0, set_temperature, process_postproc, main
from .__version__ import __version__
from .circuit import Circuit

__all__ = ['new_op', 'new_dc', 'new_tran', 'new_ac', 'new_pss',
           'new_symbolic', 'new_pz', 'queue', 'run', 'new_x0',
           'get_op_x0', 'set_temperature', 'main', 'Circuit']



# -*- coding: iso-8859-1 -*-
# switch.py
# Implementation of the voltage controlled switch model
# Copyright 2013 Giuseppe Venturini
#
# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""
Implementation of a voltage controlled switch.

This module defines two classes: switch_device, switch_model

"""

# sn1 o--+         +--o n1
#        |         |
#       +-+      \ o
#       |R|       \
#       +-+        +
#        |         |
# sn2 o--+         +--o n2

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import math

from . import options
from . import printing


class switch_device:

    """This is a general switch element.

    It has the following structure:

    .. image:: images/elem/switch1.svg

    |

    In ASCII for those who are consulting the documentation from the
    Python command line:

    ::

        sn1 o--+         +--o n1
               |         |
              +-+      \ o
              |R|       \\
              +-+        +
               |         |
        sn2 o--+         +--o n2

    The behavior is set by the model supplied.

    The device instance calls the following methods in the model:

    * ``get_i(ports_v, device)`` - output current
    * ``get_go(ports_v, device)`` - ouput conductance
    * ``get_gm(ports_v, device)`` - output transconductance
    * ``get_dc_guess(self, is_on)`` - guesses for OP

    The device instance accesses the following attributes:
    ``part_id`` (a string), the device label.

    """

    def __init__(self, n1, n2, sn1, sn2, model, ic=None, part_id='S'):
        """
        **Parameters:**

        n1 : str
            Positive output node (+)
        n2 : str
            Negative output node (-)
        sn1 : str
            Positive input node (+)
        sn2 : str
            Negative input node (-)
        model : model obj
            An instance of (v)switch_model
        ic : bool, optional
            The initial conditions: ``True`` stands for on, ``False`` for off.

        Selected methods:

        - :func:`get_output_ports` -> (n1, n2)
        - :func:`get_drive_ports` -> (n1, n2), (ns1, ns2)

        """
        class dev_class:
            pass
        self.device = dev_class()
        self.device.is_on = ic if ic is not None else False
        self.sn1 = sn1
        self.sn2 = sn2
        self.n1 = n1
        self.n2 = n2
        self.ports = ((self.n1, self.n2), (self.sn1, self.sn2))
        self.model = model
        self.opdict = {}
        self.opdict.update({'state': (float('nan'), float('nan'))})
        self.part_id = part_id
        self.is_nonlinear = True
        self.is_symbolic = True
        self.dc_guess = self.model.get_dc_guess(self.device.is_on)

    def get_drive_ports(self, op):
        """Get the ports that drive the output ports.

        **Parameters:**

        op : op solution
            The OP where the drive ports are used.

        **Returns:**

        pts : tuple of tuples of ports nodes, as: ``(port0, port1, port2 ... )``

        Where each port is in the form: ``port0 = (nplus, nminus)``
        """
        return self.ports

    def get_output_ports(self):
        """Get the output port.

        The output port is ``(n1, n2)`` for the voltage-controlled switch case.

        **Returns:**

        pts : tuple of tuples of ports nodes
            Such as: ``(port0, port1, port2 ... )``.
            Where each port is in the form: ``port0 = (nplus, nminus)``
        """
        return ((self.n1, self.n2),)

    def __str__(self):
        rep = self.model.name + " " + str(self.device.is_on)
        return rep

    def i(self, op_index, ports_v, time=0):
        """Returns the current flowing in the element.

        The element is assumed to be biased with the voltages
        applied as specified in the ``ports_v`` vector.

        **Parameters:**

        op_index : int
            The index of the output port for which the current is evaluated.
        ports_v : tuple
            A tuple constructed such as ``(voltage_across_port0, voltage_across_port1, ... )``
        time : float, optional
            The simulation time at which the evaluation is performed. It is
            needed by time-variant elements, and it has no effect here. Set it
            to ``None`` during DC analysis.

        **Returns:**

        i : int
            The output current.
        """
        ret = self.model.get_i(ports_v, self.device)
        # This may be used for debugging
        # print str(ports_v)+" Isw: %g\tRo: %g\tgm: %g" % (ret, 1/self.g(0,
        # ports_v, 0), self.g(0, ports_v, 1))
        return ret

    def update_status_dictionary(self, ports_v):
        """Updates an internal dictionary that can then be used to provide
        information to the user regarding the status of the element.

        Normally, one would call :func:`get_op_info`.

        **Returns:**

        ``None``.
        """
        if self.opdict is None:
            self.opdict = {}
        if not (self.opdict['state'] == ports_v[0] and 'R' in self.opdict):
            self.opdict['state'] = ports_v[0]
            self.opdict['R'] = float(1.0 / self.g(0, ports_v[0], 0))
            self.opdict['I'] = float(self.i(0, ports_v[0]))
            self.opdict['STATUS'] = self.device.is_on

    def get_op_info(self, ports_v):
        """Information regarding the Operating Point (OP)

        **Parameters:**

        ports_v : list of lists
            The parameter is to be set to ``[[v]]``, where ``v`` is the voltage
            applied to the switch terminals.

        **Returns:**

        op_keys : list of strings
            The labels corresponding to the numeric values in ``op_info``.
        op_info : list of floats
            The values corresponding to ``op_keys``.
        """
        self.update_status_dictionary(ports_v)
        status = "ON" if self.opdict['STATUS'] else "OFF"
        op_keys = ['Part ID', 'STATUS', "VO [V]", "VS [V]", u"R [\u2126]",
                   "I [A]"]
        op_info = [self.part_id, status, float(self.opdict['state'][0]),
                   float(self.opdict['state'][1]), self.opdict["R"],
                   self.opdict['I']]
        return op_keys, op_info

    def g(self, op_index, ports_v, port_index, time=0):
        """Returns the differential (trans)conductance.

        The transconductance is computed wrt the port specified by
        ``port_index`` when the element has the voltages specified in
        ``ports_v`` across its ports, at (simulation) ``time``.

        **Parameters:**

        ports_v : list
            Voltages applied to the switch. The list should be in the form:
            ``[voltage_across_port0, voltage_across_port1, ... ]``
        port_index : int
            The index of the output port.
        time : float
            The simulation time at which the evaluation is performed. Set it to
            ``None`` during DC analysis.

        **Returns:**

        g : float
            The transconductance.
        """

        assert op_index == 0
        assert port_index < 2

        if port_index == 0:
            return self.model.get_go(ports_v, self.device)
        if port_index == 1:
            return self.model.get_gm(ports_v, self.device)
        else:
            raise Exception("Unknown port index passed to switch: bug")

    def get_value_function(self, identifier):
        def get_value(self):
            return self.opdict[identifier]
        return get_value

    def get_netlist_elem_line(self, nodes_dict):
        """Return a netlist line corresponding to the switch."""
        return "%s %s %s %s %s %s %s" % (self.part_id, nodes_dict[self.n1],
                                nodes_dict[self.n2], nodes_dict[self.sn1],
                                nodes_dict[self.sn2], self.model.name, \
                                str(self.device.is_on))


VT_DEFAULT = 0.0
VH_DEFAULT = 0.0
RON_DEFAULT = 1.
ROFF_DEFAULT = 1. / options.gmin


class vswitch_model:
    """Voltage-controlled switch model.

    ::

        sn1 o--+         +--o n1
               |         |
              +-+      \ o
              |R|       \\
              +-+        +
               |         |
        sn2 o--+         +--o n2


    Note that:

    * R is infinite.
    * The voltage needed to close the switch is:
      :math:`V(s_{n1})-V(s_{n2}) > V_T+V_H`.
    * To re-open it, one needs to satisfy the relationship:
      :math:`V(s_{n1})-V(s_{n2}) < V_T-V_H`.

    The switch commutes between two statuses:

    * :math:`R_{OUT} = R_{OFF}`
    * :math:`R_{OUT} = R_{ON}`

    None of which can be set to zero or infinite.

    The switching characteristics are modeled with :math:`tanh(x)`.
    """

    def __init__(self, name, VT=None, VH=None, VON=None, VOFF=None, RON=None, ROFF=None):
        self.name = name
        # convert to VT and VH
        if VON is not None or VOFF is not None:
            VT, VH = self._get_VTVH_from_VONVOFF(float(VON), float(VOFF))
        self.VT = float(VT) if VT is not None else VT_DEFAULT
        self.VH = float(VH) if VH is not None else VH_DEFAULT
        self.RON = float(RON) if RON is not None else RON_DEFAULT
        self.ROFF = float(ROFF) if ROFF is not None else ROFF_DEFAULT
        self.A = (self.RON - self.ROFF) / 2
        self.B = (self.RON + self.ROFF) / 2.
        self.is_on = False
        self._set_status(self.is_on)
        self.SLOPE = 1e2

    def _get_VTVH_from_VONVOFF(self, VON, VOFF):
        if VON is None or VOFF is None:
            raise ValueError
        VT = (VON - VOFF) / 2.0 + VOFF
        return VT, VT * 1e-3

    def _get_V(self, is_on):
        """Get the effective switching voltage (hyst taken into account)
        """
        return self.VT + self.VH * 2 * (not is_on) - self.VH

    def _set_status(self, is_on):
        """Set the switch status, which meeans setting the effective
        switching voltage self.V (w hyst taken into account)
        """
        self.V = self.VT + self.VH * 2 * (not is_on) - self.VH

    def _update_status(self, vin, dev, debug=False):
        """Check the switch status and move to the other if needed.
        """
        Vtest = self._get_V(dev.is_on)
        R1 = self.A * math.tanh((vin - Vtest) * self.SLOPE) + self.B
        Vtest = self._get_V(not dev.is_on)
        R2 = self.A * math.tanh((vin - Vtest) * self.SLOPE) + self.B
        self._set_status(dev.is_on)
        if vin > self.V and not dev.is_on and R1 - R2 == 0.0:
            if debug:
                print("Switching ON: %g" % (vin,))
            dev.is_on = True
            self._set_status(dev.is_on)
        if vin < self.V and dev.is_on and R1 - R2 == 0.0:
            if debug:
                print("Switching OFF: %g" % (vin,))
            dev.is_on = False
            self._set_status(dev.is_on)
        self.is_on = dev.is_on

    def get_dc_guess(self, is_on):
        """Returns a list of two floats to be used as initial guesses for the OP analysis
        """
        return [self.VT * (.9 + is_on * .2)] * 2

    def print_model(self):
        """All the internal parameters of the model get printed out,
        for visual inspection.
        """
        arr = []
        arr.append(
            [self.name, "", "", "SWITCH MODEL", "", "", "", "", "",  "", "", ""])
        arr.append(["VT", "[V]", self.VT, "VH", "[V]:", self.VH,
                   "RON", "[ohm]", self.RON, "ROFF", "[ohm]", self.ROFF])
        print(printing.table(arr))

    def get_i(self, xxx_todo_changeme, dev, debug=False):
        """Returns the output current.
        """
        (vout, vin) = xxx_todo_changeme
        self._update_status(vin, dev)
        R = self.A * math.tanh((vin - self.V) * self.SLOPE) + self.B
        return vout / R

    def get_go(self, xxx_todo_changeme1, dev, debug=False):
        """Returns the output conductance d(I)/d(Vn1-Vn2)."""
        (vout, vin) = xxx_todo_changeme1
        self._update_status(vin, dev)
        R = self.A * math.tanh((vin - self.V) * self.SLOPE) + self.B
        return 1. / R

    def get_gm(self, xxx_todo_changeme2, dev, debug=False):
        """Returns the source to output transconductance or d(I)/d(Vsn1-Vsn2)."""
        (vout, vin) = xxx_todo_changeme2
        self._update_status(vin, dev)
        gm = self.A * self.SLOPE * (math.tanh(self.SLOPE * (self.V - vin)) ** 2 - 1) / (
            self.A * math.tanh(self.SLOPE * (self.V - vin)) - self.B) ** 2
        return gm + options.gmin




#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
# ahkab.py
# The frontend of the simulator
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""

Introduction
============

This is the core module of the simulator. It provides helper functions to save
you the need to call directly the functions in most submodules.

Do you have a circuit?
======================

To run a simulation, you'll need a circuit first: a circuit can be described
with a simulation deck or with a circuit object.

Define your circuit by means of a Circuit object
------------------------------------------------

In a Python script, describing the circuit through the
:class:`ahkab.circuit.Circuit` interface is a very versatile a choice.

Refer to :class:`ahkab.circuit.Circuit` for a complete description of the
process and the documentation of several helper functions to assist you in this
task.

You may then jump to :ref:`create-simulation-object`.

Define your circuit by means of a netlist file
----------------------------------------------

The circuit description can also be provided as a text file, also known as
netlist deck, for historical reason. This file will also typically include
simulation and post-processing directives, such as plotting.

The netlist should be described according to the rules in
:doc:`help/Netlist-Syntax`.

If you have a netlist (simulation deck) available, you have several possibilities.

The first, assuming your netlist defines some simulation would be to run it:

* you may call ``ahkab`` from the command line. The command line interface is
  described in :doc:`help/Command-Line-Help`.
* you may call :func:`main` directly from Python. Running the simulation through
  :func:`main` function allows to process the result in Python.

Alternatively, you may parse the netlist through
:func:`ahkab.netlist_parser.parse_circuit`, which will return the circuit
instance, all the simulations defined in the deck and all the post-processing
directives as well.

You may now modify the circuit and simulation objects as you please, or create
new ones, as well as run them as described in the :ref:`run-it` section.

.. _create-simulation-object:

How to create a simulation object
=================================

Next, you need to have a simulation object you would like to run.

The following functions are available to quickly create a simulation object:

.. autosummary::
    new_ac
    new_dc
    new_op
    new_pss
    new_pz
    new_symbolic
    new_tran

Click on one of the above hyperlinks to be taken to the corresponding
documentation section.

.. note::

    The functions above allow you to specify an output file. This is due to two
    main reasons:

    * Saving to a file allows you to keep a copy of the simulation results,
      which you can then inspect at a later time.
    * Simulation results may take an uncomfortably large amount of memory. The
      approach we take is that we save everything to file, and only load the
      data to memory when the user actually accesses it.

    In order for the latter to work when no output file is specified, ``ahkab``
    stores the simulation data in a temporary file provided by your OS. When the
    user exits the Python interpreter (or IPython or debugger), the file is
    removed.

.. _run-it:

Run it!
=======

Once you have a circuit and one or more simulations, it's time to run them!

The following methods are available to do so:

.. autosummary::
    run
    queue

The :func:`run` function will return the results in dictionary form.

Extras
======

The core module also contains a few extra methods which were deemed important
enough to be inserted here.

In particular, the :func:`get_op_x0` method allows the user to quickly compute
an operating point to be used to specify the linearization point for a more
complex analysis and :func:`icmodified_x0` allows the user to modify said
operating point to take into account the user-specified initial conditions in
the circuit description.

Lastly, :func:`set_temperature` can be used to quickly set the simulation
temperature.

All methods in alphabetical order
==================================

"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import atexit
import copy
import os
import sys
import tempfile

from optparse import OptionParser

import numpy as np
import scipy as sp
import sympy
import tabulate
try:
    import matplotlib
    plotting_available = True
except ImportError:
    plotting_available = False

# analyses
from . import dc_analysis
from . import transient
from . import ac
from . import pss
from . import symbolic
from . import pz

# parser
from . import netlist_parser

# misc
from . import options
from . import constants
from . import utilities

# post-processing
from . import fourier

# data display
from . import plotting
from . import printing

#py3 compat
from .py3compat import text_type

from .__version__ import __version__

global _queue, _x0s, _print

_queue = []
_print = False
_x0s = {None: None}

def new_op(guess=None, x0=None, outfile=None, verbose=0):
    """Assembles an OP analysis and returns the analysis object.

    The analysis itself can then be run with: ``ahkab.run(...)``
    or queued with ``ahkab.queue(...)`` and then run subsequently.

    **Parameters:**

    guess : boolean, optional
        if set to True, the analysis will start from an initial guess,
        hopefully speeding up the convergence of stiff circuits.

    x0 : matrix, optional
        In alternative to the ``guess`` option above, one can provide
        an explicit starting point to the OP algorithm, setting x0 to an opportunely sized
        ``numpy`` array. FIXME mention help method here
        If both x0 and guess are set, x0 takes the precedence.

    outfile : string, optional
        the filename of the output file where the results will be written.
        ``.opinfo`` is automatically added at the end to prevent different
        analyses from overwriting each-other's results.
        If unset or set to None, defaults to ``stdout``, if the simulator was
        called from the command line, otherwise, if the simulator is run from an
        interactive session, a temporary file will be used to store the data.

    verbose : int, optional
        the verbosity level, from 0 (silent, default) to 6 (debug).

    **Returns:**

    an : dict
        the analysis description

    .. seealso:: :func:`run`, :func:`queue`
    """
    if guess is None:
        guess = options.dc_use_guess
    if outfile is None or outfile == 'stdout':
        if options.cli:
            outfile = 'stdout'
        else:
            tmpfile = tempfile.NamedTemporaryFile(suffix='.op', delete=False)
            outfile = tmpfile.name
            tmpfile.close()
            atexit.register(os.remove, outfile)
    else:
        outfile += '.op'
    return {'type': 'op', 'guess': guess, 'x0': x0, 'outfile': outfile, 'verbose': verbose}


def new_dc(start, stop, points, source, sweep_type='LINEAR', guess=True, x0=None,
        outfile=None, verbose=0):
    """Assembles a DC sweep analysis and returns the analysis object.

    The analysis itself can be run with: ``ahkab.run(...)``
    or queued and then run subsequently.

    **Parameters:**

    start : float
        the start value for the sweep.

    stop : float
        the stop value for the sweep (included in the sweep points).

    points : int
        the number of sweep points.

    source : string
        the ``part_id`` of the independent current or voltage source to be swept.

    sweep_type : string, optional
        can be set to either ``options.dc_lin_step`` (linear stepping) or
        ``options.dc_log_step`` (log10 stepping). Defaults to linear.

    guess : boolean, optional
        if set to ``True``, the analysis will start from an initial guess,
        hopefully speeding up the convergence of particularly stiff circuits.

    x0 : ``numpy`` array, optional
        if the ``guess`` option above is not used, one can provide
        a starting point directly, setting ``x0`` to an opportunely sized
        ``numpy`` array.
        If both ``x0`` and ``guess`` are set, ``x0`` takes the precedence.

    outfile : string, optional
        the filename of the output file where the results will be written.
        '.dc' is automatically added at the end to prevent different
        analyses from overwriting each-other's results.
        If unset or set to ``None``, defaults to ``stdout``, if the simulator
        was called from the command line, otherwise, if the simulator is run
        from an interactive session, a temporary file will be used to store the
        data.

    verbose : int, optional
        the verbosity level, from 0 (silent, default) to 6 (debug).

    **Returns:**

    an : dict
        the analysis description

    .. seealso:: :func:`run`, :func:`queue`
    """
    if outfile is None or outfile == 'stdout':
        if options.cli:
            outfile = 'stdout'
        else:
            tmpfile = tempfile.NamedTemporaryFile(suffix='.dc', delete=False)
            outfile = tmpfile.name
            tmpfile.close()
            atexit.register(os.remove, outfile)
    else:
        outfile += '.dc'
    return {
        'type': 'dc', 'start': float(start), 'stop': float(stop), 'step': float(stop - start) / float(points - 1),
        'source': source, 'x0': x0, 'outfile': outfile, 'guess': guess, 'sweep_type': sweep_type,
        'verbose': verbose}


def new_tran(tstart, tstop, tstep, x0='op', method=transient.TRAP,
        use_step_control=True, outfile=None, verbose=0):
    """Assembles a TRAN analysis and returns the analysis object.

    The analysis itself can be run with ``ahkab.run(...)``
    or queued with ``ahkab.queue(...)`` and then run subsequently.

    **Parameters:**

    tstart : float
        the start time for the transient analysis.

    tstop : float
        the stop time.

    tstep :float
        the time step. If the step control is active, this is the
        minimum time step value that will be allowed during simulation.

    x0 : ``numpy`` array, optional
        the optional initial conditions point, :math:`x0 = x(t=0)`.

    method : string , optional
        the differentiation method to be used. Can be set to
        'IMPLICIT_EULER', 'TRAP', 'GEAR4', 'GEAR5' or 'GEAR6'.
        It defaults to 'TRAP'.

    use_step_control : boolean, optional
        Whether ste control should be enabled or not. if set to ``False``, the
        differentiation method will use a fixed time step equal to ``tstep``.

    outfile : string, optional
        the filename of the output file where the results will be written.
        '.tran' is automatically added at the end to prevent different
        analyses from overwriting each-other's results.
        If unset or set to ``None``, defaults to ``stdout``, if the simulator
        was called from the command line, otherwise, if the simulator is run
        from an interactive session, a temporary file will be used to store the
        data.

    verbose : int, optional
        the verbosity level, from 0 (silent, default) to 6 (debug).

    **Returns:**

    an : dict
        the analysis description

    .. seealso:: :func:`run`, :func:`queue`
    """
    if outfile is None or outfile == 'stdout':
        if options.cli:
            outfile = 'stdout'
        else:
            tmpfile = tempfile.NamedTemporaryFile(suffix='.tran', delete=False)
            outfile = tmpfile.name
            tmpfile.close()
            atexit.register(os.remove, outfile)
    else:
        outfile += '.tran'
    return {"type": "tran", "tstart": tstart, "tstop": tstop, "tstep": tstep,
            "method": method, "use_step_control": use_step_control, 'x0': x0,
            'outfile': outfile, 'verbose': verbose}


def new_ac(start, stop, points, x0='op', sweep_type='LOG', outfile=None, verbose=0):
    """Assembles an AC analysis and returns the analysis object.

    The analysis itself can be run with ``ahkab.run(...)``
    or queued with ``ahkab.queue(...)`` and then run subsequently.

    **Parameters:**

    start : float
        the start angular frequency, :math:`\omega _{start}`.

    stop : float
        the stop angular frequency, :math:`\omega _{stop}` (included in the
        sweep).

    points : float
        the number of points to be used the discretize the
        `[start, stop]` interval.

    x0 : string or ndarray, optional
        The linearization point for the AC analysis. If set to 'op' (default),
        the latest Operating point analysis will be used. Otherwise, you may
        supply your own linearization point in ndarray format.

    sweep_type : string, optional
        It can be set to either ``options.ac_lin_step`` (linear stepping) or
        ``options.ac_log_step`` (log10 stepping). Defaults to logarithmic
        stepping.

    outfile : string, optional
        the filename of the output file where the results will be written.
        '.ac' is automatically added at the end to prevent different
        analyses from overwriting each-other's results.
        If unset or set to ``None``, defaults to ``stdout``, if the simulator
        was called from the command line, otherwise, if the simulator is run
        from an interactive session, a temporary file will be used to store the
        data.

    verbose : int, optional
        the verbosity level, from 0 (silent, default) to 6 (debug).

    **Returns:**

    an : dict
        the analysis object (a dict)

    .. seealso:: :func:`run`, :func:`queue`
    """
    if outfile is None or outfile == 'stdout':
        if options.cli:
            outfile = 'stdout'
        else:
            tmpfile = tempfile.NamedTemporaryFile(suffix='.ac', delete=False)
            outfile = tmpfile.name
            tmpfile.close()
            atexit.register(os.remove, outfile)
    else:
        outfile += '.ac'
    return {
        'type': 'ac', 'start': start, 'stop': stop, 'points': points,
        'sweep_type': sweep_type, 'x0': x0, 'outfile': outfile,
        'verbose': verbose}


def new_pss(period, x0=None, points=None, method=options.BFPSS, autonomous=False,
            outfile=None, verbose=0):
    """Assembles a Periodic Steady State (PSS) analysis and returns the analysis object.

    The analysis itself can be run with: ``ahkab.run(...)`` or queued with
    ``ahkab.queue(...)`` and then run subsequently.

    **Parameters:**

    period : float
        the time period of the solution, in seconds. This value is required,
        autonomous circuits are currently unsupported.

    x0 : ``numpy`` array, optional
        the starting point solution, used at :math:`t=0`.

    points : int, optional
        the number of points to use to discretize the PSS solution. If not set,
        if method is 'shooting', defaults to ``options.shooting_default_points``

    method : string, optional
        The method to be employed to attempt a PSS solution of the circuit.
        It can be either ``ahkab.BFPSS`` or ``ahkab.SHOOTING``.

    autonomous : bool, optional
        Whether the circuit is autonomous or not.
        Non-autonomous circuits are currently unsupported!

    mna, Tf, D : ``numpy`` arrays, optional
        The matrices to be used to solve the circuit.
        They are optional, if they have already been computed, reusing them saves time.

    outfile : string, optional
        The filename of the output file where the results will be written.
        '.tran' is automatically added at the end to prevent different
        analyses from overwriting each-other's results.
        If unset or set to ``None``, defaults to ``stdout``, if the simulator
        was called from the command line, otherwise, if the simulator is run
        from an interactive session, a temporary file will be used to store the
        data.

    verbose : int, optional
        The verbosity level, from 0 (silent, default) to 6 (debug).

    **Returns:**

    an : dict
        the analysis object (a dict)

    .. seealso:: :func:`run`, :func:`queue`
    """
    if outfile is None or outfile == 'stdout':
        if options.cli:
            outfile = 'stdout'
        else:
            tmpfile = tempfile.NamedTemporaryFile(suffix='.'+method.lower(), delete=False)
            outfile = tmpfile.name
            tmpfile.close()
            atexit.register(os.remove, outfile)
    else:
        outfile += '.' + method.lower()
    return {
        'type': "pss", "method": method, 'period': period, 'points': points,
        'autonomous': autonomous, 'x0': x0, 'outfile': outfile, 'verbose': verbose}


def new_pz(input_source=None, output_port=None, shift=0.0, MNA=None, outfile=None,
           x0='op', verbose=0):
    """Assembles a Pole-Zero analysis and returns the analysis object.

    The analysis itself can be run with: ``ahkab.run(...)`` or queued with
    ``ahkab.queue(...)`` and then run subsequently.

    **Parameters:**

    input_source : str or instance
        the input source for zero calculation

    output_port : tuple or single node
        the output port. If it is composed of only one node, then the
        second node is assumed to be GND.

    shift : float, optional
        Perform the calculation at a shifted freq ``shift``.

    MNA : ndarray, optional
        the numpy matrix to be used to solve the circuit.
        It is optional, but, if it's already been computed,
        reusing it will save time.

    outfile : string, optional
        The filename of the output file where the results will be written.
        '.pz' is automatically added at the end to prevent different
        analyses from overwriting each-other's results.
        If unset or set to ``None``, defaults to ``stdout``, if the simulator
        was called from the command line, otherwise, if the simulator is run
        from an interactive session, a temporary file will be used to store the
        data.

    x0 : ``numpy`` array or str, optional
        the optional linearization point. If set to a string, it must be
        the result of an .OP analysis (use ``'op'``) or an .IC condition
        defined in the netlist. It has no effect on linear circuits.

    verbose : int, optional
        The verbosity level, from 0 (silent, default) to 6 (debug).

    **Returns:**

    an : the analysis description object, a dict instance.
    """
    if outfile is None or outfile == 'stdout':
        if options.cli:
            outfile = 'stdout'
        else:
            tmpfile = tempfile.NamedTemporaryFile(suffix='.pz', delete=False)
            outfile = tmpfile.name
            tmpfile.close()
            atexit.register(os.remove, outfile)
    else:
        outfile += '.pz'
    return {'type': "pz", 'input_source':input_source, 'x0':x0,
            'output_port':output_port, 'shift':shift, 'MNA':MNA,
            'outfile':outfile, 'verbose':verbose}


def new_symbolic(source=None, ac_enable=True, r0s=False, subs=None, outfile=None, verbose=0):
    """Assembles a Symbolic analysis and returns the analysis object.

    The analysis itself can be run with ``ahkab.run(...)``
    or queued with ``ahkab.queue(...)`` and then run subsequently.

    **Parameters:**

    source : string, optional
        if ``source`` is set, the transfer function between the current or
        voltage source ``source`` and each circuit unknown will be
        evaluated, with symbolic evaluation of DC gain, poles and zeros.
        ``source`` is to be set to the ``part_id`` of an independent current
        or voltage source in the circuit, eg. ``'V1'`` or ``'Iin'``. This
        computation should be avoided for large circuit, as indiscriminate
        transfer function, gain and singularities evaluation in large
        circuits can result in very long run times and needs a significant
        amount of RAM, on top of an already resource intensive symbolic
        analysis.  We suggest manually evaluating selected transfer
        functions of interest instead.

    ac_enable : bool, optional
        If set to ``True`` (default), the frequency-dependent elements will
        be considered, otherwise the algorithm will focus on
        low frequency solutions, where all capacitors are replaced with open
        circuits and all inductors are short circuits, usually providing a much
        easier circuit.

    r0s : bool, optional
        If set to ``True``, the finite output conductances of transistors
        ``go`` (where :math:`go = 1/r_0`) will be taken into account,
        otherwise they will be considered infinite (default).
        The finite output conductances generally introduce a significant
        additional complexity in large circuits, sometimes of interest to the
        designer, sometimes simply introducing 2nd and 3rd order effects of
        little-to-no interest, which would produce no significant contribution
        in a numerical analysis, but come at a high computation price in a
        symbolic analysis.
        A possible approach in those cases may be disabling this option and
        explicitly introducing additional conductances where deemed of interest.

    subs : dict, optional
        ``subs`` is a dictionary of substitutions to be performed before
        attempting to solve the circuit. For example, if two
        resistances ``R1`` and ``R2`` are to be equal, set ``subs={'R2':'R1'}``
        and ``R1`` will be replaced by an instance of ``R2``. This may
        simplify the solution (or allow finding one in reasonable
        time for complex circuits).

    outfile : string, optional
        The filename of the output file where the results will be written.
        '.symbolic' is automatically added at the end to prevent different
        analyses from overwriting each-other's results.
        If unset or set to ``None``, defaults to ``stdout``, if the simulator
        was called from the command line, otherwise, if the simulator is run
        from an interactive session, a temporary file will be used to store the
        data.

    verbose : int, optional
        The verbosity level, from 0 (silent, default) to 6 (debug).

    **Returns:**

    an : dict
        the analysis description

    .. seealso:: :func:`run`, :func:`queue`
    """
    if outfile is None or outfile == 'stdout':
        if options.cli:
            outfile = 'stdout'
        else:
            tmpfile = tempfile.NamedTemporaryFile(suffix='.symbolic', delete=False)
            outfile = tmpfile.name
            tmpfile.close()
            atexit.register(os.remove, outfile)
    else:
        outfile += '.symbolic'
    return {
        'type': "symbolic", 'source': source, 'ac_enable': ac_enable, 'r0s': r0s, 'subs': subs,
        'outfile': outfile, 'verbose': verbose}


def queue(*analysis):
    """Queue one or more analyses to execute them subsequently with :func:`run`.

    **Parameters**

    analysis : one or more analysis descriptions.
        The analyses to be queued.

    **Returns:**

    None
    """
    global _queue
    for an in analysis:  # let's hope the user knows what he's doing!
        _queue += [copy.deepcopy(an)]


def run(circ, an_list=None):
    """Run analyses on a circuit.

    **Parameters:**

    circ : circuit instance
        The circuit to be simulated.

    an_queue : list, optional
        the list of analyses to be performed. If unset, it defaults to those
        queued with ``queue``.

    **Returns:**

    results : dict
        the results (in dict form)

    .. seealso:: :func:`queue`
    """
    results = {}

    if not an_list:
        an_list = _queue
    else:
        an_list = copy.deepcopy(an_list)
        if type(an_list) == tuple:
            an_list = list(an_list)
        elif type(an_list) == dict:
            an_list = [an_list] # run(mycircuit, op1)

    while len(an_list):
        an_item = an_list.pop(0)
        an_type = an_item.pop('type')
        if 'x0' in an_item and isinstance(an_item['x0'], text_type):
            printing.print_warning("%s has x0 set to %s, unavailable. Using 'None'." %
                                   (an_type.upper(), an_item['x0']))
            an_item['x0'] = None
        r = analysis[an_type](circ, **an_item)
        results.update({an_type: r})
        if an_type == 'op':
            _x0s.update({'op': r})
            _x0s.update({'op+ic': icmodified_x0(circ, r)})
            _handle_netlist_ics(circ, an_list, ic_list=[])
    return results


def new_x0(circ, icdict):
    """Builds an ``x0`` matrix from user supplied values.

    Supplying a custom x0 can be useful:
    - To aid convergence in tough circuits.
    - To start a transient simulation from a particular x0

    **Parameters:**

    circ : circuit instance
        The circuit
    icdict : dict
        a dictionary specifying the node voltages and branch currents,
        where appropriate, in V and A, respectively, assembled as shown
        in the next section.
        All unspecified node voltages default to ``0`` V and all
        unspecified currents default to 0.

    The user-specified values are to be provided as follows:

    - to specify a nodal voltage: ``{'V(node)':<voltage value>}``
    - to specify a branch current: ``'I(<element>)':<current value>}``


    Examples:

    - ``{'V(n1)':2.3, 'V(n2)':0.45, ...}``
    - ``{'I(L1)':1.03e-3, I(V4):2.3e-6, ...}``

    .. note::
        This simulator uses the normal convention, also known as the
        `Passive sign convention
        <https://en.wikipedia.org/wiki/Passive_sign_convention>`_.

    **Returns:**

    x0 : ``numpy`` array
        The assembled x0.
    """

    return dc_analysis.build_x0_from_user_supplied_ic(circ, icdict)


def icmodified_x0(circ, x0):
    """Modify ``x0`` to take into account the ICs in the circuit.

    **Parameters:**

    circ : circuit instance
        The circuit instance from which the initial conditions are to be
        extracted.

    x0 : ``numpy`` array
        The vector to which the initial conditions are to be applied.
    """
    return dc_analysis.modify_x0_for_ic(circ, x0)


def get_op_x0(circ):
    """Shorthand to specify and run an OP analysis to get a linearization
    point.

    **Parameters:**

    circ : circuit instance
        The circuit instance for which the linearization point is sought.

    **Returns:**

    x0 : an OP solution object
        The linearization point.
    """
    return run(circ, [new_op()])


def set_temperature(T):
    """Set the simulation temperature, in Celsius."""
    T = float(T)
    if T > 300:
        printing.print_warning("The temperature will be set to %f \xB0 C.")
    constants.T = utilities.Celsius2Kelvin(T)


def process_postproc(postproc_list, title, results, outfilename):
    """Runs the post-processing operations, such as plotting.

    Not meant for end users.

    deprecated in 0.10

    **Parameters:**

    postproc_list : list,
        list of post processing operations

    title : string
        the deck title

    results: dict
        the results to be plotted (which may include including ones that are not needed too).

    outfilename: string
        if the plots are saved to disk, this is the filename without extension
    """
    show_plots = False
    skip_plots = False
    for postproc in postproc_list:
        if postproc['type'] == "plot" and outfilename == 'stdout':
            skip_plots = True
            printing.print_warning("Plotting and printing the results to " +
                                   "stdout are incompatible options. " +
                                   "Plotting skipped.")
            break
    for index, postproc in enumerate(postproc_list):
        if postproc['type'] == "plot" and not skip_plots:
            plotting.plot_results(title, postproc["l2l1"],
                                  results[postproc["analysis"]],
                                  "%s-%d.%s" % (outfilename, index,
                                                options.plotting_outtype))
            show_plots = True
        elif postproc['type'] == "four":
            for v in postproc['variables']:
                f, F, THD = fourier.fourier(v, results['tran'],
                                            postproc['fund'])
                printing.print_fourier(v, f, F, THD, outfile='stdout')
        elif postproc['type'] == "fft":
            postproc.pop('type') # get rid of spurious entry
            f, F, THD = fourier.spicefft(tran_results=results['tran'],
                                         **postproc)
            printing.print_spicefft(postproc['label'], f, F, THD,
                                    uformat=postproc['uformat'],
                                    window=postproc['window'],
                                    outfile=outfilename +
                                    '.lis'*(outfilename!='stdout'))
    if show_plots and options.plotting_show_plots:
        plotting.show_plots()

analysis = {'op': dc_analysis.op_analysis, 'dc': dc_analysis.dc_analysis,
            'tran': transient.transient_analysis, 'ac': ac.ac_analysis,
            'pss': pss.pss_analysis, 'symbolic': symbolic.symbolic_analysis,
            'temp': set_temperature, 'pz':pz.calculate_singularities}


def main(filename, outfile="stdout", verbose=3):
    """Method to call ahkab from a Python script with a netlist file.

    **Parameters:**

    filename : string
        The netlist filename.

    outfile : string, optional
        The outfiles base name, the suffixes shown below will be added.
        With the exception of the magic value ``stdout`` which causes
        ahkab to print out instead of to disk.

    verbose : int, optional
        the verbosity level, from 0 (silent) to 6 (debug).
        It defaults to 3, the same as running ahkab through its command
        line interface.

    Filename suffixes, for each analysis:

    - Alternate Current (AC): ``.ac``
    - Direct Current (DC): ``.dc``
    - Operating Point (OP): ``.opinfo``
    - Periodic Steady State (PSS): ``.pss``
    - Pole-zero Analysis (PZ): ``.pz``
    - TRANsient (TRAN): ``.tran``
    - Symbolic: ``.symbolic``

    **Returns:**

    res : dict
        A dictionary containing the computed results.
    """
    printing.print_info_line(
        ("This is ahkab %s running with:" % (__version__), 6), verbose)
    printing.print_info_line(
        ("  Python %s" % (sys.version.split('\n')[0],), 6), verbose)
    printing.print_info_line(("  Numpy %s" % (np.__version__), 6), verbose)
    printing.print_info_line(("  Scipy %s" % (sp.__version__), 6), verbose)
    printing.print_info_line(("  Sympy %s" % (sympy.__version__), 6), verbose)
    printing.print_info_line(("  Tabulate %s" % (tabulate.__version__), 6), verbose)
    if plotting_available:
        printing.print_info_line(("  Matplotlib %s" % (matplotlib.__version__),
                                  6), verbose)
        printing.print_info_line(("  -> backend: %s" %
                                 (matplotlib.get_backend()), 6), verbose)
        printing.print_info_line(("  -> matplotlibrc: %s" %
                                 (matplotlib.matplotlib_fname()), 6), verbose)
    else:
        printing.print_info_line(
            ("  Matplotlib not found.", 6), verbose)


    read_netlist_from_stdin = (filename is None or filename == "-")
    (circ, directives, postproc_direct) = netlist_parser.parse_circuit(
        filename, read_netlist_from_stdin)

    printing.print_info_line(("Checking circuit for common mistakes...", 6),
                             verbose, print_nl=False)
    check, reason = utilities.check_circuit(circ)
    if not check:
        printing.print_general_error(reason)
        sys.exit(3)
    printing.print_info_line(("done.", 6), verbose)

    if verbose > 3 or _print:
        print("Parsed circuit:")
        print(circ)
        print("Models:")
        for m in circ.models:
            circ.models[m].print_model()
        print("")

    ic_list = netlist_parser.parse_ics(directives)
    _handle_netlist_ics(circ, an_list=[], ic_list=ic_list)
    results = {}
    for an in netlist_parser.parse_analysis(circ, directives):
        if 'outfile' not in list(an.keys()) or not an['outfile']:
            an.update(
                {'outfile': outfile + ("." + an['type']) * (outfile != 'stdout')})
        if 'verbose' in list(an.keys()) and (an['verbose'] is None or an['verbose'] < verbose) \
           or not 'verbose' in list(an.keys()):
            an.update({'verbose': verbose})
        _handle_netlist_ics(circ, [an], ic_list=[])
        if verbose >= 4:
            printing.print_info_line(("Requested an.:", 4), verbose)
            printing.print_analysis(an)
        results.update(run(circ, [an]))

    postproc_list = netlist_parser.parse_postproc(circ, postproc_direct)
    if len(postproc_list) > 0 and len(results):
        process_postproc(postproc_list, circ.title, results, outfile)

    return results


def _handle_netlist_ics(circ, an_list, ic_list):
    for ic in ic_list:
        ic_label = list(ic.keys())[0]
        icdict = ic[ic_label]
        _x0s.update({ic_label: new_x0(circ, icdict)})
    for an in an_list:
        if 'x0' in an and isinstance(an['x0'], text_type):
            if an['x0'] in list(_x0s.keys()):
                an['x0'] = _x0s[an['x0']]
            elif an_list.index(an) == 0:
                raise ValueError(("The x0 '%s' is not available." % an["x0"]) +\
                                 (an['x0'] == 'op' or an['x0'] == 'op+ic')*
                                 " Perhaps you forgot to define an .OP?")




# -*- coding: iso-8859-1 -*-
# time_functions.py
# Time functions for independent sources
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.
"""
This module contains several basic time functions.

The classes that are found in module are useful to provide a time-varying
characteristic to independent sources.

Notice that the time functions are not restricted to those provided here, the
user is welcome to provide his own.
Implementing a custom time function is easy and common practice, as long as you
are interfacing to the simulator through Python. Please see the dedicated section
:ref:`define-custom-time-functions` below.

Classes defined in this module
------------------------------

.. autosummary::
    pulse
    pwl
    sin
    exp
    sffm
    am

Supplying a time function to an independent source
--------------------------------------------------

Providing a time-dependent characteristic to an independent source is very
simple and probably best explained with an example.

Let's say we wish to define a sinusoidal voltage source with no offset,
amplitude 5V and 1kHz frequency.

It is done in two steps:

* first we define the time function with the built-in class
  :class:`ahkab.time_functions.sin`:

    .. code-block:: python

        sin1k = time_functions.sin(vo=0, va=5, freq=1e3)

* Then we define the voltage source and we assign the time function to it:

    .. code-block:: python

        cir.add_vsource('V1', 'n1', cir.gnd, 1, function=mys)

In the example above, the sine wave is assigned to a voltage source ``'V1'``,
that gets added to a circuit ``cir`` (not shown).

.. _define-custom-time-functions:

Defining custom time functions
------------------------------

Defining a custom time function is easy, all you need is either:

* A function that takes a ``float`` (the time) and returns the function
  value,
* An instance with a ``__call__(self, time)`` method. This solution
  allows having internal parameters, typically set through the constructor.

In both cases, in time-based simulations, the simulator will call the object at
every time step, supplying a single parameter, the simulation time (``time`` in
the following, of type ``float``).

In turn, the simulator expects to receive as return value a ``float``,
corresponding to the value of the time-dependent function at the time specified
by the ``time`` variable.

If the time-dependent function is used to define the characteristics of a
voltage source (:class:`VSource`), its return value has to be expressed in Volt.
In the case of a current source (:class:`ISource`), the return value is to be
expressed in Ampere.

The standard notation applies.

As an example, we'll define a custom time-dependent voltage source, having a
:math:`\\mathrm{sinc}(ft)` characteristic. In this example, :math:`f` has a
value of 10kHz.

First we define the time function, in this case we'll do that through the Python
``lambda`` construct.

.. code-block:: python

    mys = lambda t: 1 if not t else math.sin(math.pi*1e4*t)/(math.pi*1e4*t)

Then, we define the circuit -- a very simple one in this case -- and assign our
``mys`` function to ``V1``. In the following circuit, we simply apply the
voltage from ``V1`` to a resistor ``R1``.

.. code-block:: python

    import ahkab
    cir = ahkab.Circuit('Test custom time functions')
    cir.add_resistor('R1', 'n1', cir.gnd, 1e3)
    cir.add_vsource('V1', 'n1', cir.gnd, 1, function=mys)
    tr = ahkab.new_tran(0, 1e-3, 1e-5, x0=None)
    r = ahkab.run(cir, tr)['tran']

Plotting ``Vn1`` and the expected result (:math:`\\mathrm{sinc}(ft)`) we
get:

.. plot::

    import math
    import numpy as np
    import pylab
    import ahkab
    cir = ahkab.Circuit('Test custom time functions')
    cir.add_resistor('R1', 'n1', cir.gnd, 1e3)
    mys = lambda t: 1 if not t else math.sin(math.pi*1e4*t)/(math.pi*1e4*t)
    cir.add_vsource('V1', 'n1', cir.gnd, 1, function=mys)
    tr = ahkab.new_tran(0, 1e-3, 1e-5, x0=None)
    r = ahkab.run(cir, tr)['tran']
    t = r.get_x()
    pylab.hold(True)
    pylab.plot(t, r['vn1'], 'o', ms=3, label='V(n1) (simulation)')
    npsin1k = np.frompyfunc(mys, 1, 1)
    pylab.plot(t, npsin1k(t), label='sinc(ft) (theory)')
    pylab.legend()
    pylab.xlabel('t [s]')
    pylab.ylabel('Voltage [V]')


Module reference
----------------
"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import math

from scipy.interpolate import InterpolatedUnivariateSpline

time_fun_specs = {'sin': { #VO VA FREQ TD THETA
    'tokens': ({
               'label': 'vo',
               'pos': 0,
               'type': float,
               'needed': True,
               'dest': 'vo',
               'default': None
               },
               {
               'label': 'va',
               'pos': 1,
               'type': float,
               'needed': True,
               'dest': 'va',
               'default': None
               },
               {
               'label': 'freq',
               'pos': 2,
               'type': float,
               'needed': True,
               'dest': 'freq',
               'default': None
               },
               {
               'label': 'td',
               'pos': 3,
               'type': float,
               'needed': False,
               'dest': 'td',
               'default': 0.
               },
               {
               'label': 'theta',
               'pos': 4,
               'type': float,
               'needed': False,
               'dest': 'theta',
               'default': 0
               }
               )
        },'exp': { #EXP(V1 V2 TD1 TAU1 TD2 TAU2)
    'tokens': ({
               'label': 'v1',
               'pos': 0,
               'type': float,
               'needed': True,
               'dest': 'v1',
               'default': None
               },
               {
               'label': 'v2',
               'pos': 1,
               'type': float,
               'needed': True,
               'dest': 'v2',
               'default': None
               },
               {
               'label': 'td1',
               'pos': 2,
               'type': float,
               'needed': False,
               'dest': 'td1',
               'default': 0.
               },
               {
               'label': 'tau1',
               'pos': 3,
               'type': float,
               'needed': True,
               'dest': 'tau1',
               'default': None
               },
               {
               'label': 'td2',
               'pos': 4,
               'type': float,
               'needed': False,
               'dest': 'td2',
               'default': float('inf')
               },
               {
               'label': 'tau2',
               'pos': 5,
               'type': float,
               'needed': False,
               'dest': 'tau2',
               'default': float('inf')
               }
               )
        },'pulse': { #PULSE(V1 V2 TD TR TF PW PER)
    'tokens': ({
               'label': 'v1',
               'pos': 0,
               'type': float,
               'needed': True,
               'dest': 'v1',
               'default': None
               },
               {
               'label': 'v2',
               'pos': 1,
               'type': float,
               'needed': True,
               'dest': 'v2',
               'default': None
               },
               {
               'label': 'td',
               'pos': 2,
               'type': float,
               'needed': False,
               'dest': 'td',
               'default': 0.
               },
               {
               'label': 'tr',
               'pos': 3,
               'type': float,
               'needed': True,
               'dest': 'tr',
               'default': None
               },
               {
               'label': 'tf',
               'pos': 4,
               'type': float,
               'needed': True,
               'dest': 'tf',
               'default': None
               },
               {
               'label': 'pw',
               'pos': 5,
               'type': float,
               'needed': True,
               'dest': 'pw',
               'default': None
               },
               {
               'label': 'per',
               'pos': 6,
               'type': float,
               'needed': True,
               'dest': 'per',
               'default': None
               })
        }, 'sffm': { ## SFFM(VO VA FC MDI FS TD)
    'tokens': ({
               'label': 'vo',
               'pos': 0,
               'type': float,
               'needed': True,
               'dest': 'vo',
               'default': None
               },
               {
               'label': 'va',
               'pos': 1,
               'type': float,
               'needed': True,
               'dest': 'va',
               'default': None
               },
               {
               'label': 'fc',
               'pos': 2,
               'type': float,
               'needed': False,
               'dest': 'fc',
               'default': None
               },
               {
               'label': 'mdi',
               'pos': 3,
               'type': float,
               'needed': True,
               'dest': 'mdi',
               'default': None
               },
               {
               'label': 'fs',
               'pos': 4,
               'type': float,
               'needed': True,
               'dest': 'fs',
               'default': None
               },
               {
               'label': 'td',
               'pos': 5,
               'type': float,
               'needed': False,
               'dest': 'td',
               'default': 0.
               })
        }, 'am': { #AM(sa oc fm fc [td])
    'tokens': ({
               'label': 'sa',
               'pos': 0,
               'type': float,
               'needed': True,
               'dest': 'sa',
               'default': None
               },
               {
               'label': 'oc',
               'pos': 1,
               'type': float,
               'needed': True,
               'dest': 'oc',
               'default': None
               },
               {
               'label': 'fm',
               'pos': 2,
               'type': float,
               'needed': True,
               'dest': 'fm',
               'default': None
               },
               {
               'label': 'fc',
               'pos': 3,
               'type': float,
               'needed': True,
               'dest': 'fc',
               'default': None
               },
               {
               'label': 'td',
               'pos': 4,
               'type': float,
               'needed': False,
               'dest': 'td',
               'default': None
               })
        }
}

#
# Functions for time dependent sources  #
#

class pulse(object):
    """Square wave aka pulse function

    .. image:: images/elem/pulse.svg

    **Parameters:**

    v1 : float
        Square wave low value.
    v2 : float
        Square wave high value.
    td : float
        Delay time to the first ramp, in seconds. Negative values are considered
        as zero.
    tr : float
        Rise time in seconds, from the low value ``v1`` to the pulse high value
        ``v2``.
    tf : float
        Fall time in seconds, from the pulse high value ``v2`` to the low value
        ``v1``.
    pw : float
        Pulse width in seconds.
    per : float
        Periodicity interval in seconds.
    """
    # PULSE(V1 V2 TD TR TF PW PER)

    def __init__(self, v1, v2, td, tr, pw, tf, per):
        self.v1 = v1
        self.v2 = v2
        self.td = max(td, 0.0)
        self.per = per
        self.tr = tr
        self.tf = tf
        self.pw = pw
        self._type = "V"

    def __call__(self, time):
        """Evaluate the pulse function at the given time."""
        if time is None:
            time = 0
        time = time - self.per * int(time / self.per)
        if time < self.td:
            return self.v1
        elif time < self.td + self.tr:
            return self.v1 + ((self.v2 - self.v1) / (self.tr)) * (time - self.td)
        elif time < self.td + self.tr + self.pw:
            return self.v2
        elif time < self.td + self.tr + self.pw + self.tf:
            return self.v2 + ((self.v1 - self.v2) / (self.tf)) * (time - (self.td + self.tr + self.pw))
        else:
            return self.v1

    def __str__(self):
        return "type=pulse " + \
            self._type.lower() + "1=" + str(self.v1) + " " + \
            self._type.lower() + "2=" + str(self.v2) + \
            " td=" + str(self.td) + " per=" + str(self.per) + \
            " tr=" + str(self.tr) + " tf=" + str(self.tf) + \
            " pw=" + str(self.pw)


class sin(object):
    """Sine wave

    .. image:: images/elem/sin.svg

    Mathematically, the sine wave function is defined as:

    * :math:`t < t_d`:

    .. math::

        f(t) = v_o + v_a \\sin\\left(\\pi \\phi/180 \\right)

    * :math:`t \\ge t_d`:

    .. math::

        f(t) = v_o + v_a \\exp\\left[-(t - t_d)\,\\theta \\right] \\sin\\left[2 \\pi f (t - t_d) + \\pi \\phi/180\\right]

    **Parameters:**

    vo : float
        Offset value.
    va : float
        Amplitude.
    freq : float
        Sine frequency in Hz.
    td : float, optional
        time delay before beginning the sinusoidal time variation, in seconds. Defaults to 0.
    theta : float optional
        damping factor in 1/s. Defaults to 0 (no damping).
    phi : float, optional
        Phase delay in degrees. Defaults to 0 (no phase delay).

    .. note::

        This implementation is consistent with the SPICE simulator, other simulators use
        different formulae.

    """
    # SIN(VO VA FREQ TD THETA)

    def __init__(self, vo, va, freq, td=0., theta=0., phi=0.):
        self.vo = vo
        self.va = va
        self.freq = freq
        self.td = td
        self.theta = theta
        self.phi = phi
        self._type = "V"

    def __call__(self, time):
        """Evaluate the sine function at the given time."""
        if time is None:
            time = 0
        if time < self.td:
            return self.vo + self.va*math.sin(math.pi*self.phi/180.)
        else:
            return self.vo + self.va * math.exp((self.td - time)*self.theta) \
                   * math.sin(2*math.pi*self.freq*(time - self.td) + \
                              math.pi*self.phi/180.)

    def __str__(self):
        return "type=sin " + \
            self._type.lower() + "o=" + str(self.vo) + " " + \
            self._type.lower() + "a=" + str(self.va) + \
            " freq=" + str(self.freq) + " theta=" + str(self.theta) + \
            " td=" + str(self.td)


class exp(object):
    """Exponential wave

    .. image:: images/elem/exp.svg

    Mathematically, it is described by the equations:

    * :math:`0 \\le t < TD1`:

    .. math::

        f(t) = V1

    * :math:`TD1 < t < TD2`

    .. math::

        f(t) = V1+(V2-V1) \\cdot \\left[1-\\exp
               \\left(-\\frac{t-TD1}{TAU1}\\right)\\right]

    * :math:`t > TD2`

    .. math::

        f(t) = V1+(V2-V1) \\cdot \\left[1-\\exp
               \\left(-\\frac{t-TD1}{TAU1}\\right)\\right]+(V1-V2) \\cdot
               \\left[1-\\exp \\left(-\\frac{t-TD2}{TAU2}\\right)\\right]

    **Parameters:**

    v1 : float
        Initial value.
    v2 : float
        Pulsed value.
    td1 : float
        Rise delay time in seconds.
    td2 : float
        Fall delay time in seconds.
    tau1 : float
        Rise time constant in seconds.
    tau2 : float
        Fall time constant in seconds.
    """
    # EXP(V1 V2 TD1 TAU1 TD2 TAU2)

    def __init__(self, v1, v2, td1, tau1, td2, tau2):
        self.v1 = v1
        self.v2 = v2
        self.td1 = td1
        self.tau1 = tau1
        self.td2 = td2
        self.tau2 = tau2
        self._type = "V"

    def __call__(self, time):
        """Evaluate the exponential function at the given time."""
        if time is None:
            time = 0
        if time < self.td1:
            return self.v1
        elif time < self.td2:
            return self.v1 + (self.v2 - self.v1) * \
                   (1 - math.exp(-1*(time - self.td1)/self.tau1))
        else:
            return self.v1 + (self.v2 - self.v1) * \
                   (1 - math.exp(-1*(time - self.td1)/self.tau1)) + \
                   (self.v1 - self.v2)*(1 - math.exp(-1*(time - self.td2)/self.tau2))

    def __str__(self):
        return "type=exp " + \
            self._type.lower() + "1=" + str(self.v1) + " " + \
            self._type.lower() + "2=" + str(self.v2) + \
            " td1=" + str(self.td1) + " td2=" + str(self.td2) + \
            " tau1=" + str(self.tau1) + " tau2=" + str(self.tau2)


class sffm(object):
    """Single-Frequency FM (SFFM) waveform

    .. image:: images/elem/fm.svg

    Mathematically, it is described by the equations:

    * :math:`0 \\le t \\le t_D`:

    .. math::

        f(t) = V_O

    * :math:`t > t_D`

    .. math::

        f(t) = V_O + V_A \\cdot \\sin \\left[2\\pi f_C (t - t_D) + MDI
               \\sin \\left[2 \\pi f_S (t - t_D) \\right] \\right]

    **Parameters:**

    vo : float
        Offset in Volt or Ampere.
    va : float
        Amplitude in Volt or Ampere.
    fc : float
        Carrier frequency in Hz.
    mdi : float
        Modulation index.
    fs : float
        Signal frequency in HZ.
    td : float
        Time delay before the signal begins, in seconds.
    """
    # SFFM(VO VA FC MDI FS)

    def __init__(self, vo, va, fc, mdi, fs, td):
        self.vo = vo
        self.va = va
        self.fc = fc
        self.mdi = mdi
        self.fs = fs
        self.td = td
        self._type = "V"

    def __call__(self, time):
        """Evaluate the SFFM function at the given time."""
        if time is None:
            time = 0
        if time <= self.td:
            return self.vo
        else:
            return self.vo + self.va*math.sin(2*math.pi*self.fc*(time - self.td) +
                                              self.mdi*math.sin(2*math.pi*self.fs*
                                                                (time - self.td))
                                              )

    def __str__(self):
        return "type=sffm vo=%g va=%g fc=%g mdi=%g fs=%g td=%g" % \
                (self.vo, self.va, self.fc, self.mdi, self.fs, self.td)

class am(object):
    """Amplitude Modulated (AM) waveform

    .. image:: images/elem/am.svg

    Mathematically, it is described by the equations:

    * :math:`0 \\le t \\le t_D`:

    .. math::

        f(t) = O

    * :math:`t > t_D`

    .. math::

        f(t) = SA \\cdot \\left[OC + \\sin \\left[2\\pi f_m (t - t_D) \\right]
               \\right] \\cdot \\sin \\left[2 \\pi f_c (t - t_D) \\right]

    **Parameters:**

    sa : float
        Signal amplitude in Volt or Ampere.
    fc : float
        Carrier frequency in Hertz.
    fm : float
        Modulation frequency in Hertz.
    oc : float
        Offset constant, setting the absolute magnitude of the modulation.
    td : float
        Time delay before the signal begins, in seconds.
    """
    # AM(sa oc fm fc <td>)

    def __init__(self, sa, fc, fm, oc, td):
        self.sa = sa
        self.fc = fc
        self.fm = fm
        self.oc = oc
        self.td = td
        self._type = "V"

    def __call__(self, time):
        """Evaluate the AM function at the given time."""
        if time is None:
            time = 0
        if time <= self.td:
            return 0.
        else:
            return self.sa*(self.oc + math.sin(2*math.pi*self.fm*
                                               (time - self.td)))* \
                   math.sin(2*math.pi*self.fc*(time - self.td))

    def __str__(self):
        return "type=am sa=%g oc=%g fm=%g fc=%g td=%g" % \
                (self.sa, self.oc, self.fm, self.fc, self.td)

class pwl(object):
    """Piece-Wise Linear (PWL) waveform

    .. image:: images/elem/pwl.svg

    A piece-wise linear waveform is defined by a sequence of points
    :math:`(x_i, y_i)`.

    Please supply the abscissa values :math:`\\{x\\}_i` in the vector
    ``x``, the ordinate values :math:`\\{y\\}_i` in the vector ``y``,
    separately.


    **Parameters:**

    x : sequence-like
        The abscissa values of the interpolation points.
    y : sequence-like
        The ordinate values of the interpolation points.
    repeat : boolean, optional
        Whether the waveform should be repeated after its end. If set to
        ``True``, ``repeat_time`` also needs to be set to define when the
        repetition begins. Defaults to ``False``.
    repeat_time : float, optional
        In case the waveform is set to be repeated, setting the ``repeat`` flag
        above, the parameter, defined in seconds, set the first time instant at
        which the waveform repetition happens.
    td : float, optional
        Time delay before the signal begins, in seconds. Defaults to zero.

    **Example:**

    The following code::

        import ahkab
        import numpy as np
        import pylab as plt
        # vs = (x1, y1, x2, y2, x3, y3 ...)
        vs = (60e-9, 0, 120e-9, 0, 130e-9, 5, 170e-9, 5, 180e-9, 0)
        x, y = vs[::2], vs[1::2]
        fun = ahkab.time_functions.pwl(x, y, repeat=1, repeat_time=60e-9, td=0)
        myg = np.frompyfunc(fun, 1, 1)
        t = np.linspace(0, 5e-7, 2000)
        plt.plot(t, myg(t), lw=3)
        plt.xlabel('Time [s]'); plt.ylabel('Arbitrary units []')


    Produces:

    .. plot::

        import ahkab
        import numpy as np
        import pylab as plt
        vs = (60e-9, 0, 120e-9, 0, 130e-9, 5, 170e-9, 5, 180e-9, 0)
        x, y = vs[::2], vs[1::2]
        fun = ahkab.time_functions.pwl(x, y, repeat=1, repeat_time=60e-9, td=0)
        myg = np.frompyfunc(fun, 1, 1)
        t = np.linspace(0, 5e-7, 2000)
        plt.figure(figsize=(6, 3)); plt.grid()
        plt.plot(t, myg(t), lw=3)
        plt.xlabel('Time [s]'); plt.ylabel('Arbitrary units []')
        plt.tight_layout()


    """

    def __init__(self, x, y, repeat=False, repeat_time=0, td=0):
        self.x = x
        self.y = y
        self.repeat = repeat
        self.repeat_time = repeat_time
        if self.repeat_time == max(x):
            self.repeat_time = 0
        self.td = td
        self._type = "V"
        self._f = InterpolatedUnivariateSpline(self.x, self.y, k=1)

    def __call__(self, time):
        """Evaluate the PWL function at the given time."""
        time = self._normalize_time(time)
        return self._f(time)

    def _normalize_time(self, time):
        if time is None:
            time = 0
        if time <= self.td:
            time = 0
        elif time > self.td:
            time = time - self.td
            if self.repeat:
                if time > max(self.x):
                    time = (time - max(self.x)) % \
                           (max(self.x) - self.repeat_time) + \
                           self.repeat_time
                else:
                    pass
        return time

    def __str__(self):
        pwl_str = "type=pwl"
        tv = " "
        for x, y in zip(self.x, self.y):
            tv += "%g %g "
        pwl_str += tv
        if self.td:
            pwl_str += "td=%g " % self.td
        if self.repeat:
            pwl_str = "RPT=%g " % self.repeat_time
        return pwl_str[:-1]




# -*- coding: utf-8 -*-
# options.py
# Global options file
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""This module contains options and configuration switches the user
may tune to meet his needs.

The default values are sensible options for the general case.
"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import os

import numpy as np

try:
    from matplotlib import rcParams as plotting_rcParams
except ImportError:
    plotting_rcParams = {}

#: Encoding of the netlist files.
encoding = 'utf8'

#: Cache size to be used in :func:`ahkab.utilities.memoize`, defaults to 512MB
cache_len = 67108864 # 512MB

#: A boolean to differentiate command line execution from module import
#: When cli is False, no printing and no weird stdout stuff.
cli = False

# global: errors
#: Voltage absolute tolerance.
vea = 1e-6
#: Voltage relative tolerance.
ver = 1e-3
#: Current absolute tolerance.
iea = 1e-9
#: Current relative tolerance.
ier = 1e-3

# global: circuit
#: Minimum conductance to ground.
gmin = 1e-12
#: Should we show to the user results pertaining to nodes introduced by
#: components or by the simulator?
print_int_nodes = True

# global: solving
#: Dense matrix limit: if the dimensions of the square MNA matrix are bigger,
#: use sparse matrices.
dense_matrix_limit = 400
#: Should we damp artificially the first NR iterations? See also
#: :func:`ahkab.dc_analysis.get_td`.
nr_damp_first_iters = False
#: In all NR iterations, lock the nodes controlling non-linear elements. See
#: also :func:`ahkab.dc_analysis.get_td`.
nl_voltages_lock = True     # Apply damping - slows down solution.
#: Non-linear nodes lock factor:
#: if we allow the voltage on controlling ports to change too much, we may
#: have current/voltage overflows. Think about the diode characteristic.
#: So we allow them to change of ``nl_voltages_lock_factor``
#: :math:`\cdot V_{th}` at most and damp all variables accordingly.
nl_voltages_lock_factor = 4

#: Whether the standard solving method can be used.
use_standard_solve_method = True
#: Whether the gmin-settping homothopy can be used.
use_gmin_stepping = True
#: Whether the source-stepping homothopy can be used.
use_source_stepping = True

#: When printing out to the user, whether we can suppress trailing zeros.
print_suppress = False
#: When printing out to the user, how many decimal digits to show at maximum.
print_precision = np.get_printoptions()['precision']

# dc
#: Maximum allowed NR iterations during a DC analysis.
dc_max_nr_iter = 10000
#: Enable guessing to init the NR solver during a DC analysis.
dc_use_guess = True
#: Do not perform an init DC guess if its effort is higher than
#: this value.
dc_max_guess_effort = 250000
# shorthand to set logarithmic stepping in DC analyses.
dc_log_step = 'LOG'
# shorthand to set linear stepping in DC analyses.
dc_lin_step = 'LIN'
#: Can we skip troublesome points during DC sweeps?
dc_sweep_skip_allowed = True

# transient
#: The default differentiation method for transient analyses.
default_tran_method = "TRAP"
#: Minimum allowed discretization step for time.
hmin = 1e-20
#: Maximum number of time iterations for transient analyses
#: Notice the default (0) means no limit is enforced.
transient_max_time_iter = 0  # disabled
#: Maximum number of NR iterations for transient analyses.
transient_max_nr_iter = 20
#: In a transisent analysis, if a prediction value is avalilable,
#: use it as first guess for ``x(n+1)``, otherwise ``x(n)`` is used.
transient_prediction_as_x0 = True
#: Use aposteriori step control?
transient_use_aposteriori_step_control = True
#: Step change threshold:
#: we do not want to redo the iteraction if the aposteriori check suggests a step
#: that is very close to the one we already used. A value of 0.9 seems to be a
#: good idea.
transient_aposteriori_step_threshold = 0.9
#: Disable all step control in transient analyses.
transient_no_step_control = False
#: Minimum capacitance to ground.
cmin = 1e-18

# pss
BFPSS = 'brute-force'
SHOOTINGPSS = 'shooting'

# shooting
#: Default number of points for a shooting analysis.
shooting_default_points = 100
#: Maximum number of NR iterations for shooting analyses.
shooting_max_nr_iter = 10000

# bfpss
#: Default number of points for a BFPSS analysis.
bfpss_default_points = 100
#: Maximum number of NR iterations for BFPSS analyses.
bfpss_max_nr_iter = 10000

# symbolic
#: Enable the manual solver:
#: solve the circuit equations one at a time as you might do "manually".
symb_sympy_manual_solver = False
#: Formulate the equations with conductances and at the last moment
#: swap resistor symbols back in. It seems to make sympy play nicer.
#: Sometimes.
symb_formulate_with_gs = False

# ac
ac_log_step = 'LOG'
ac_lin_step = 'LIN'
#: Maximum number of NR iterations for AC analyses.
ac_max_nr_iter = 20
#: Use degrees instead of rads in AC phase results.
ac_phase_in_deg = False

#pz
#: Maximum considered angular frequency in rad/s for PZ analyses.
pz_max = 1e12

# plotting
# Set to None to disable writing plots to disk
#: Should plots be shown to the user? This variable is set to ``True``
#: automatically if a screen is detected in Unix systems.
#:
#: Notice that by default ahkab both shows plots *and* saves them to disk.
plotting_show_plots = ('DISPLAY' in os.environ)
#: Wait for the user to close the plot? If set to ``False``, plots are created
#: and immediately destroyed.
plotting_wait_after_plot = True
#: Format to be used when writing plots to disk.
plotting_outtype = "png"
#: Matplotlib line plot style: see matplotlib's doc.
plotting_style = "-o"
#: Plotting line width.
plotting_lw = 1.25
#: Default size for plots showed to the user, in inches.
plotting_display_figsize = (12.94, 8)
#: Default size for plots saved to disk.
plotting_save_figsize = (20, 10)
# plotting_rcParams['font.family'] = 'Baskerville'
plotting_rcParams['axes.labelsize'] = 11
plotting_rcParams['xtick.labelsize'] = 11
plotting_rcParams['ytick.labelsize'] = 11
plotting_rcParams['legend.fontsize'] = 11

# Windows
RECT_WINDOW = 'RECT'
BART_WINDOW = 'BART'
HANN_WINDOW = 'HANN'
HAMM_WINDOW = 'HAMM'
BLACK_WINDOW = 'BLACK'
HARRIS_WINDOW = 'HARRIS'
GAUSS_WINDOW = 'GAUSS'
KAISER_WINDOW = 'KAISER'
WINDOWS_NAMES = dict(RECT='RECTANGULAR', BART='BARTLETT',
                     HANN='HANNING', HAMM='HAMMING',
                     BLACK='BLACKMAN-HARRIS', HARRIS='HARRIS',
                     GAUSS='GAUSSIAN', KAISER='KAISER')




# -*- coding: iso-8859-1 -*-
# shooting.py
# Shooting analysis module
# Copyright 2009 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""Periodic steady state analysis based on the shooting method."""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import numpy as np
import numpy.linalg

from . import transient
from . import implicit_euler
from . import dc_analysis
from . import ticker
from . import options
from . import circuit
from . import printing
from . import utilities
from . import results
from . import devices


def shooting_analysis(circ, period, step=None, x0=None, points=None, autonomous=False,
             matrices=None, outfile='stdout', vector_norm=lambda v: max(abs(v)), verbose=3):
    """Performs a periodic steady state analysis based on the algorithm described in:

        Brambilla, A.; D'Amore, D., "Method for steady-state simulation of
        strongly nonlinear circuits in the time domain," *Circuits and
        Systems I: Fundamental Theory and Applications, IEEE Transactions on*,
        vol.48, no.7, pp.885-889, Jul 2001.

        http://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=933329&isnumber=20194

    The results have been computed again by me, the formulation is not exactly the
    same, but the idea behind the shooting algorithm is.

    This method allows us to have a period with many points without having to
    invert a huge matrix (and being limited to the maximum matrix size).

    A transient analysis is performed to initialize the solver.

    We compute the change in the last point, calculating several matrices in
    the process.
    From that, with the same matrices we calculate the changes in all points,
    starting from 0 (which is the same as the last one), then 1, ...

    Key points:

    - Only non-autonomous circuits are supported.
    - The time step is constant.
    - Implicit Euler is used as DF.

    **Parameters:**

    circ : Circuit instance
        The circuit description class.
    period : float
        The period of the solution.
    step : float, optional
        The time step between consecutive points.
        If not set, it will be computed from ``period`` and ``points``.
    points : int, optional
        The number of points to be used. If not set, it will be computed from
        ``period`` and ``step``.
    autonomous : bool, optional
        This parameter has to be ``False``, autonomous circuits are not
        currently supported.
    matrices : dict, optional
        A dictionary that may have as keys 'MNA', 'N' and 'D', with entries set
        to the corresponding MNA-formulation matrices, in case they have been
        already computed and the user wishes to save time by reusing them.
        Defaults to ``None`` (recompute).
    outfile : string, optional
        The output filename. Please use ``stdout`` (the default) to print to the
        standard output.
    verbose : boolean, optional
        Verbosity switch (0-6). It is set to zero (print errors only)
        if ``outfile`` == 'stdout'``, as not to corrupt the data.

    Notice that ``step`` and ``points`` are mutually exclusive options:

    - if ``step`` is specified, the number of points will be automatically determined.
    - if ``points`` is set, the step will be automatically determined.
    - if none of them is set, ``options.shooting_default_points`` will be used as points.

    **Returns:**

    sol : PSS solution object or ``None``
        The solution. If the circuit can't be solve, ``None`` is returned instead.
    """

    if outfile == "stdout":
        verbose = 0

    printing.print_info_line(("Starting periodic steady state analysis:", 3),
                             verbose)
    printing.print_info_line(("Method: shooting", 3), verbose)

    if isinstance(x0, results.op_solution):
        x0 = x0.asarray()

    # Do we have MNA T D or do we need to build them
    # MNA & T
    if (matrices is None or type(matrices) != dict or 'MNA' not in matrices or
        'Tf' not in matrices):
        # recalculate
        mna, Tf = dc_analysis.generate_mna_and_N(circ, verbose=verbose)
        mna = utilities.remove_row_and_col(mna)
        Tf = utilities.remove_row(Tf, rrow=0)
    elif not matrices['MNA'].shape[0] == matrices['Tf'].shape[0]:
        raise ValueError("MNA matrix and N vector have different number of" +
                         " rows.")
    else:
        mna, Tf = matrices['MNA'], matrices['Tf']
    # D
    if matrices is None or 'D' not in matrices or matrices['D'] is None:
        D = transient.generate_D(circ, [mna.shape[0], mna.shape[0]])
        D = utilities.remove_row_and_col(D)
    elif not mna.shape == matrices['D'].shape:
        raise ValueError("MNA matrix and D matrix have different sizes.")
    else:
        D = matrices['D']

    points, step = utilities.check_step_and_points(step, points, period,
                                                   options.shooting_default_points)

    n_of_var = mna.shape[0]
    locked_nodes = circ.get_locked_nodes()

    printing.print_info_line(("Starting TRAN analysis for algorithm init: " +
                              ("stop=%g, step=%g... " % (10*points*step, step)),
                              3), verbose, print_nl=False)
    xtran = transient.transient_analysis(circ=circ, tstart=0, tstep=step,
                                         tstop=10*points*step, method="TRAP",
                                         x0=None, mna=mna, N=Tf, D=D,
                                         use_step_control=False,
                                         outfile=outfile+".tran",
                                         return_req_dict={"points": points},
                                         verbose=0)
    if xtran is None:
        print("failed.")
        return None
    printing.print_info_line(("done.", 3), verbose)

    x = []
    for index in range(points):
        x.append(xtran[index * n_of_var:(index + 1) * n_of_var, 0])

    tick = ticker.ticker(increments_for_step=1)

    MAass_static, MBass = _build_static_MAass_and_MBass(mna, D, step)

    # This contains the time invariant part, Tf. The time variable component,
    # Tt, is always the same, since the time interval is the same this holds all
    # time-dependent sources (both V/I).

    # silly numpy sum array of size 6,1 with array of size 6
    # get array of size 6,6
    Tf = Tf.squeeze()
    Tass_static_vector = _build_Tass_static_vector(circ, Tf, points, step, tick,
                                                   n_of_var, verbose)

    converged = False
    printing.print_info_line(("Solving... ", 3), verbose, print_nl=False)
    tick.reset()
    tick.display(verbose > 2)

    iteration = 0  # newton iteration counter
    conv_counter = 0

    while True:
        dx = []
        Tass_variable_vector = []
        MAass_variable_vector = []
        for index in range(points):
            if index == 0:
                xn_minus_1 = x[points - 1]
            else:
                xn_minus_1 = x[index - 1]
            MAass_variable, Tass_variable = _get_variable_MAass_and_Tass(circ,
                    x[index], xn_minus_1, mna, D, step, n_of_var)
            MAass_variable_vector.append(MAass_variable + MAass_static)
            Tass_variable_vector.append(Tass_variable +
                                        Tass_static_vector[index])

        dxN = _compute_dxN(MAass_variable_vector, MBass, Tass_variable_vector,
                           n_of_var, points)
        td = dc_analysis.get_td(dxN.reshape(-1, 1), locked_nodes, n=-1)
        x[points - 1] = td * dxN + x[points - 1]

        for index in range(points - 1):
            if index == 0:
                dxi_minus_1 = dxN
            else:
                dxi_minus_1 = dx[index - 1]
            dx.append(_compute_dx(MAass_variable_vector[index], MBass,
                      Tass_variable_vector[index], dxi_minus_1))
            td = dc_analysis.get_td(dx[index].reshape(-1, 1), locked_nodes,
                                    n=-1)
            x[index] = td*dx[index] + x[index]
        dx.append(dxN)

        if (_vector_norm_wrapper(dx, vector_norm) < min(options.ver, options.ier) *
                _vector_norm_wrapper(x, vector_norm) + min(options.vea, options.iea)):
            # and (dc_analysis.vector_norm(residuo) <
            # options.er*dc_analysis.vector_norm(x) + options.ea):
            if conv_counter == 3:
                converged = True
                break
            else:
                conv_counter = conv_counter + 1
        elif vector_norm(dx[points - 1]) is np.nan:  # needs work fixme
            raise OverflowError
            # break
        else:
            conv_counter = 0
            tick.step()

        if options.shooting_max_nr_iter and iteration == options.shooting_max_nr_iter:
            printing.print_general_error("Hitted SHOOTING_MAX_NR_ITER (" +
                                         str(options.shooting_max_nr_iter) +
                                         "), iteration halted.")
            converged = False
            break
        else:
            iteration = iteration + 1

    tick.hide(verbose > 2)
    if converged:
        printing.print_info_line(("done.", 3), verbose)
        t = np.arange(points) * step
        t = t.reshape((1, points))
        xmat = x[0].reshape(-1, 1)
        for index in range(1, points):
            xmat = numpy.concatenate((xmat, x[index].reshape(-1, 1)), axis=1)
        sol = results.pss_solution(circ=circ, method="shooting", period=period,
                                   outfile=outfile)
        sol.set_results(t, xmat)
        # print_results(circ, x, fdata, points, step)
    else:
        print("failed.")
        sol = None
    return sol


def _vector_norm_wrapper(vector, norm_fun):
    _max = 0
    for elem in vector:
        new_max = norm_fun(elem)
        if _max < new_max:
            _max = new_max
    return _max


def _build_static_MAass_and_MBass(mna, D, step):
    (C1, C0) = implicit_euler.get_df_coeff(step)
    MAass = mna + D * C1
    MBass = np.dot(D, C0)
    return (MAass, MBass)


def _build_Tass_static_vector(circ, Tf, points, step, tick, n_of_var, verbose=3):
    Tass_vector = []
    nv = circ.get_nodes_number()
    printing.print_info_line(("Building Tass...", 5), verbose, print_nl=False)

    tick.reset()
    tick.display(verbose > 2)
    for index in range(0, points):
        Tt = numpy.zeros((n_of_var,))
        v_eq = 0
        time = index * step
        for elem in circ:
            if (isinstance(elem, devices.VSource) or
                isinstance(elem, devices.ISource)) and elem.is_timedependent:
                # time dependent source
                if isinstance(elem, devices.VSource):
                    Tt[nv - 1 + v_eq] = -1.0 * elem.V(time)
                elif isinstance(elem, devices.ISource):
                    if elem.n1:
                        Tt[elem.n1 - 1] = Tt[elem.n1 - 1] + elem.I(time)
                    if elem.n2:
                        Tt[elem.n2 - 1] = Tt[elem.n2 - 1] - elem.I(time)
            if circuit.is_elem_voltage_defined(elem):
                v_eq = v_eq + 1
        tick.step()
        Tass_vector.append(Tf + Tt)
    tick.hide(verbose > 2)
    printing.print_info_line(("done.", 5), verbose)

    return Tass_vector


def _get_variable_MAass_and_Tass(circ, xi, xi_minus_1, M, D, step, n_of_var):
    Tass = np.zeros((n_of_var,))
    J = np.zeros((n_of_var, n_of_var))
    (C1, C0) = implicit_euler.get_df_coeff(step)

    for elem in circ:
        # build all dT(xn)/dxn (stored in J) and T(x)
        if elem.is_nonlinear:
            output_ports = elem.get_output_ports()
            for index in range(len(output_ports)):
                n1, n2 = output_ports[index]
                ports = elem.get_drive_ports(index)
                v_ports = []
                for port in ports:
                    v = 0  # build v: remember we trashed the 0 row and 0 col of mna -> -1
                    if port[0]:
                        v = v + xi[port[0] - 1]
                    if port[1]:
                        v = v - xi[port[1] - 1]
                    v_ports.append(v)
                if n1:
                    Tass[n1 - 1] = Tass[n1 - 1] + elem.i(index, v_ports)
                if n2:
                    Tass[n2 - 1] = Tass[n2 - 1] - elem.i(index, v_ports)
                for pindex in range(len(ports)):
                    if n1:
                        if ports[pindex][0]:
                            J[n1 - 1, ports[pindex][0] - 1] = \
                                J[n1 - 1, ports[pindex][0] - 1] + \
                                elem.g(index, v_ports, pindex)
                        if ports[pindex][1]:
                            J[n1 - 1, ports[pindex][1] - 1] = \
                                J[n1 - 1, ports[pindex][1] - 1] - \
                                elem.g(index, v_ports, pindex)
                    if n2:
                        if ports[pindex][0]:
                            J[n2 - 1, ports[pindex][0] - 1] = \
                                J[n2 - 1, ports[pindex][0] - 1] - \
                                elem.g(index, v_ports, pindex)
                        if ports[pindex][1]:
                            J[n2 - 1, ports[pindex][1] - 1] = \
                                J[n2 - 1, ports[pindex][1] - 1] + \
                                elem.g(index, v_ports, pindex)

    Tass = Tass + np.dot(D*C1, xi) + np.dot(M, xi) + np.dot(D * C0, xi_minus_1)

    return J, Tass


def _compute_dxN(MAass_vector, MBass, Tass_vector, n_of_var, points):
    temp_mat1 = np.eye(n_of_var)
    for index in range(points):
        temp_mat1 = -np.linalg.solve(MAass_vector[index], np.dot(MBass, temp_mat1))
    temp_mat2 = np.zeros((n_of_var,))
    for index in range(points):
        temp_mat3 = -np.linalg.solve(MAass_vector[index], Tass_vector[index])
        for index2 in range(index + 1, points):
            temp_mat3 = -np.linalg.solve(MAass_vector[index2], np.dot(MBass, temp_mat3))
        temp_mat2 = temp_mat2 + temp_mat3

    dxN = np.dot(np.linalg.inv(np.eye(n_of_var) - temp_mat1), temp_mat2)

    return dxN


def _compute_dx(MAass, MBass, Tass, dxi_minus_1):
    dxi = -np.linalg.solve(MAass, np.dot(MBass, dxi_minus_1) + Tass)
    # dxi = -1 * np.linalg.inv(MAass) * (MBass * dxi_minus_1 + Tass)
    return dxi



import numpy as np
import scipy
from . import options
from . import constants
from .utilities import convergence_check


def mdn_solver(x, mna, circ, T, MAXIT, nv, locked_nodes, time=None,
               print_steps=False, vector_norm=None, debug=True):
    """
    Solves a problem like F(x) = 0 using the Newton Algorithm with a variable
    damping.

    Where:

    .. math::

        F(x) = mna*x + T + T(x)

    * :math:`mna` is the Modified Network Analysis matrix of the circuit
    * :math:`T(x)` is the contribute of nonlinear elements to KCL
    * :math:`T` contains the contributions of the independent sources, time
    * invariant and linear

    If :math:`x(0)` is the initial guess, every :math:`x(n+1)` is given by:

    .. math::
        x(n+1) = x(n) + td \\cdot dx

    Where :math:`td` is a damping coefficient to avoid overflow in non-linear
    components and excessive oscillation in the very first iteration. Afterwards
    :math:`td=1` To calculate :math:`td`, an array of locked nodes is needed.

    The convergence check is done this way:

    **Parameters:**

    x : ndarray
        The initial guess. If set to ``None``, it will be initialized to all
        zeros. Specifying a initial guess may improve the convergence time of
        the algorithm and determine which solution (if any) is found if there
        are more than one.
    mna : ndarray
        The Modified Network Analysis matrix of the circuit, reduced, see above.
    circ : circuit instance
        The circuit instance.
    T : ndarray,
        The :math:`T` vector described above.
    MAXIT : int
        The maximum iterations that the method may perform.
    nv : int
        Number of nodes in the circuit (counting the ref, 0)
    locked_nodes : list of tuples
        A list of ports driving non-linear elements, generated by
        ``circ.get_locked_nodes()``
    time : float or None, optional
        The value of time to be passed to non_linear _and_ time variant
        elements.
    print_steps : boolean, optional
        Show a progress indicator, very verbose. Defaults to ``False``.
    vector_norm : function, optional
        An R^N -> R^1 function returning the norm of a vector, for convergence
        checking. Defaults to the maximum norm, ie :math:`f(x) = max(|x|)`,
    debug : int, optional
        Debug flag that will result in an array being returned containing
        node-by-node convergence information.

    **Returns:**

    sol : ndarray
        The solution.
    err : ndarray
        The remaining error.
    converged : boolean
        A boolean that is set to ``True`` whenever the method exits because of a
        successful convergence check. ``False`` whenever convergence problems
        where found.
    N : int
        The number of NR iterations performed.
    convergence_by_node : list
        If ``debug`` was set to ``True``, this list has the same size of the MNA
        matrix and contains the information regarding which nodes fail to
        converge in the circuit. Ie. ``if convergence_by_node[j] == False``,
        node ``j`` has a convergence problem. This may significantly help
        debugging non-convergent circuits.

    """
    if vector_norm is None:
        vector_norm = lambda v: max(abs(v)),
    mna_size = mna.shape[0]
    nonlinear_circuit = circ.is_nonlinear()
    if x is None:
        # if no guess was specified, its all zeros
        x = np.zeros((mna_size, 1))
    else:
        if x.shape[0] != mna_size:
            raise ValueError("x0s size is different from expected: got "
                             "%d-elements x0 with an MNA of size %d" %
                             (x.shape[0], mna_size))
    if T is None:
        #printing.print_warning(
        #    "dc_analysis.mdn_solver called with T==None, setting T=0. BUG or no sources in circuit?")
        T = np.zeros((mna_size, 1))

    sparse = mna_size > options.dense_matrix_limit
    # We allocate the matrices once and then reuse them
    if sparse:
        mna = scipy.sparse.coo_matrix(mna)
        J = scipy.sparse.lil_matrix((mna_size, mna_size))
    else:
        J = np.zeros((mna_size, mna_size))
    Tx = np.zeros((mna_size, 1))
    converged = False
    iteration = 0
    while iteration < MAXIT:  # newton iteration counter
        iteration += 1
        if nonlinear_circuit:
            # build dT(x)/dx (stored in J) and Tx(x)
            J[:, :] = 0.0
            Tx[:, 0] = 0.0
            for elem in circ:
                if elem.is_nonlinear:
                    _update_J_and_Tx(J, Tx, x, elem, time)
        residuo = mna.dot(x) + T + nonlinear_circuit*Tx

        if sparse:
            lu = scipy.sparse.linalg.splu(scipy.sparse.csc_matrix(mna + nonlinear_circuit*J))
            dx = lu.solve(-residuo)
        else:
            dx = np.linalg.solve(mna + nonlinear_circuit*J, -residuo)
        x = x + get_td(dx, locked_nodes, n=iteration) * dx
        if not nonlinear_circuit:
            converged = True
            break
        elif convergence_check(x, dx, residuo, nv - 1)[0]:
            converged = True
            break
        # if vector_norm(dx) == np.nan: #Overflow
        #   raise OverflowError
    if debug and not converged:
        # re-run the convergence check, only this time get the results
        # by node, so we can show to the users which nodes are misbehaving.
        converged, convergence_by_node = convergence_check(x, dx, residuo, 
                                                           nv - 1, debug=True)
    else:
        convergence_by_node = []
    return (x, residuo, converged, iteration, convergence_by_node)

def _update_J_and_Tx(J, Tx, x, elem, time):
    out_ports = elem.get_output_ports()
    for index in range(len(out_ports)):
        n1, n2 = out_ports[index]
        n1m1, n2m1 = n1 - 1, n2 - 1
        dports = elem.get_drive_ports(index)
        v_dports = []
        for port in dports:
            v = 0.  # build v: remember we removed the 0 row and 0 col of mna -> -1
            if port[0]:
                v = v + x[port[0] - 1, 0]
            if port[1]:
                v = v - x[port[1] - 1, 0]
            v_dports.append(v)
        if hasattr(elem, 'gstamp') and hasattr(elem, 'istamp'):
            iis, gs = elem.gstamp(v_dports, time)
            J[iis] += gs.reshape(-1)
            iis, i = elem.istamp(v_dports, time)
            Tx[iis] += i.reshape(-1)
            continue
        if n1 or n2:
            iel = elem.i(index, v_dports, time)
        if n1:
            Tx[n1m1, 0] = Tx[n1m1, 0] + iel
        if n2:
            Tx[n2m1, 0] = Tx[n2m1, 0] - iel
        for iindex in range(len(dports)):
            if n1 or n2:
                g = elem.g(index, v_dports, iindex, time)
            if n1:
                if dports[iindex][0]:
                    J[n1m1, dports[iindex][0] - 1] += g
                if dports[iindex][1]:
                    J[n1m1, dports[iindex][1] - 1] -= g
            if n2:
                if dports[iindex][0]:
                    J[n2m1, dports[iindex][0] - 1] -= g
                if dports[iindex][1]:
                    J[n2m1, dports[iindex][1] - 1] += g


def get_td(dx, locked_nodes, n=-1):
    """Calculates the damping coefficient for the Newthon method.

    The damping coefficient is choosen as the lowest between:

    - the damping required for the first NR iterations, a parameter which is set
      through the integer ``options.nr_damp_first_iters``.
    - If ``options.nl_voltages_lock`` evaluates to ``True``, the biggest damping
      factor that keeps the change in voltage across the locked nodes pairs less
      than the maximum variation allowed, set by:
      ``(options.nl_voltages_lock_factor * Vth)``
    - Unity.

    **Parameters:**

    dx : ndarray
        The undamped increment returned by the NR solver.
    locked_nodes : list
        A vector of tuples of (internal) nodes that are a port of a non-linear
        component.
    n : int, optional
        The NR iteration counter

    .. note::

        If ``n`` is set to ``-1`` (or any negative value), ``td`` is independent
        from the iteration number and ``options.nr_damp_first_iters`` is ignored.

    **Returns:**

    td : float
        The damping coefficient.

    """
    if not options.nr_damp_first_iters or n < 0:
        td = 1
    else:
        if n < 10:
            td = 1e-2
        elif n < 20:
            td = 0.1
        else:
            td = 1
    td_new = 1
    if options.nl_voltages_lock:
        for (n1, n2) in locked_nodes:
            if n1 != 0:
                if n2 != 0:
                    if abs(dx[n1 - 1, 0] - dx[n2 - 1, 0]) > options.nl_voltages_lock_factor * constants.Vth():
                        td_new = (options.nl_voltages_lock_factor * constants.Vth()) / abs(
                            dx[n1 - 1, 0] - dx[n2 - 1, 0])
                else:
                    if abs(dx[n1 - 1, 0]) > options.nl_voltages_lock_factor * constants.Vth():
                        td_new = (options.nl_voltages_lock_factor * constants.Vth()) / abs(
                            dx[n1 - 1, 0])
            else:
                if abs(dx[n2 - 1, 0]) > options.nl_voltages_lock_factor * constants.Vth():
                    td_new = (options.nl_voltages_lock_factor * constants.Vth()) / abs(
                        dx[n2 - 1, 0])
            if td_new < td:
                td = td_new
    return td



# -*- coding: iso-8859-1 -*-
# pz.py
# Pole-Zero evaluation methods
# Copyright 2014 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""
This module offers the functions needed to perform a numeric
pole-zero extraction.

Currently, this module implements the MD algorithm, more may be
added in the future.

A description of the algorithm is found in the following references:


    Haley, S.B., "The generalized eigenproblem: pole-zero computation,"
    *Proceedings of the IEEE*, vol.76, no.2, pp.103,120, Feb 1988

and:

    Raghuram, R.; Divekar, D.; Wang, P., "Implementation of pole-zero
    analysis in SPICE based on the MD method," Circuits and Systems, 1991.,
    *Proceedings of the 34th Midwest Symposium on*, pp.380,
    383 vol.1, 14-17 May 1991

Frequency sweeping -- or shifting -- is performed with a random frequency
kick, currently, hoping not to kick so hard that we end up on the negative side.
A bisection method would be better and hopefully will be implemented soon.

Overview
~~~~~~~~

Two main methods are available in this module:

 * :func:`calculate_singularities`, which computes both zeros and poles,
 * :func:`calculate_poles`, which only computes the poles.

Currently this module uses dense matrices.

Reference
~~~~~~~~~

"""
from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import copy

import numpy as np

from . import circuit
from . import dc_analysis 
from . import devices
from . import transient
from . import plotting
from . import printing
from . import py3compat
from . import options
from . import results

specs = {'pz': {'tokens': ({
                           'label': 'output',
                           'pos': 0,
                           'type': str,
                           'needed': False,
                           'dest': 'output_port',
                           'default': None
                           },
                           {
                           'label': 'input',
                           'pos': 1,
                           'type': str,
                           'needed': False,
                           'dest': 'input_source',
                           'default': None
                           },
                           {
                           'label': 'shift',
                           'pos': 2,
                           'type': float,
                           'needed': False,
                           'dest': 'shift',
                           'default': 0.0
                           })
               }
        }

def _enlarge_matrix(M):
    if M is None:
        return np.zeros((1, 1))
    else:
        return np.vstack((np.hstack((M, np.zeros((M.shape[0], 1)))),
                         np.zeros((1, M.shape[1] + 1))))

def calculate_poles(mc, MNA=None, x0=None, outfile=None, verbose=0):
    """Calculate the circuit poles.

    **Parameters:**

    mc : circuit instance
        The circuit to be analyzed.
    MNA : ndarray, optional
        The Modified Nodal Analysis matrix, if available.
        In case the circuit is non-linear, MNA should include the contributes
        of the non-linear elements (ie the Jacobian :math:`J`). 
    x0 : ndarray or op_solution, optional
         The linearization point. Only needed for non-linear circuits.
    outfile : str or None, optional
        The data filename.
    verbose : int, optional
        Verbosity level, from 0 (silent, default) to 6 (debug).

    **Returns:**

    pz_sol : pz_solution instance
        The PZ solution, with no zeros.
    """
    return calculate_singularities(mc, input_source=None, output_port=None, 
                                   MNA=None, shift=0)[0]

def calculate_singularities(mc, input_source=None, output_port=None, MNA=None,
                            x0=None, shift=0, outfile=None, verbose=0):
    """Calculate poles and zeros.

    By default, only poles are calculated, as they need no information
    other than the circuit description.

    To activate zeros calculation, it is necessary:

    * to specify an input source (``input_source``),
    * to specify an output port (``output_port``).

    **Parameters:**

    mc : circuit instance
        The circuit to be analyzed.
    input_source : string or element, optional
        If zeros are to be calculated, set this to the input surce.
    output_port : external node (ref. to gnd) or tuple of external nodes, opt
        If zeros are to be calculated, set this to the output nodes.
    MNA : ndarray, optional
        The Modified Nodal Analysis matrix, if available.
        In case the circuit is non-linear, MNA should include the contributes
        of the non-linear elements (ie the Jacobian :math:`J`). 
    x0 : ndarray or op_solution, optional
         The linearization point. Only needed for non-linear circuits.
    shift : float, optional
        Shift frequency at which the algorithm should be run.
    outfile : str or None, optional
        The data filename.
    verbose : int, optional
        Verbosity level, from 0 (silent, default) to 6 (debug).

    **Returns:**

    pz_sol : pz_solution instance
        The PZ solution

    """
    calc_zeros = (input_source is not None) and (output_port is not None)
    if calc_zeros:
        if type(input_source) not in py3compat.string_types:
            input_source = input_source.part_id
        if type(output_port) in py3compat.string_types:
            output_port = plotting._split_netlist_label(output_port)[0] 
            output_port = [o[1:] for o in output_port]
            output_port = [o.lower() for o in output_port]
            if len(output_port) == 1:
                # we refer to the ground implicitely
                output_port += ['0']
        if np.isscalar(output_port):
            output_port = (output_port, mc.gnd)
        if (type(output_port) == tuple or type(output_port) == list) \
           and type(output_port[0]) in py3compat.string_types:
            output_port = [mc.ext_node_to_int(o) for o in output_port]
        we_got_source = False
        for e in mc:
            if e.part_id == input_source:
                we_got_source = True
                break
        if not we_got_source:
            raise ValueError('Source %s not found in circuit.' % input_source)
        RIIN = []
        ROUT = []
    if MNA is None:
        MNA, N = dc_analysis.generate_mna_and_N(mc)
        if mc.is_nonlinear():
            # setup x0
            if x0 is None:
                printing.print_warning("PZ: No linearization point provided. Using x0 = 0.")
                x0 = np.zeros((MNA.shape[0] - 1, 1))
            else:
                if isinstance(x0, results.op_solution):
                    x0 = x0.asarray()
                # else
                    # hopefully x0 is an ndarray!
                printing.print_info_line(("Using the supplied op as " +
                                      "linearization point.", 5), verbose)
            J, _ = dc_analysis.build_J_and_Tx(x0, MNA.shape[0]-1, mc, time=0., sparse=False)
            MNA[1:, 1:] += J
    D = transient.generate_D(mc, MNA[1:, 1:].shape)
    MNAinv = np.linalg.inv(MNA[1:, 1:] + shift*D[1:, 1:])
    nodes_m1 = mc.get_nodes_number() - 1
    vde1 = -1
    MC = np.zeros((MNA.shape[0] - 1, 1))
    TCM = None
    dei_source = 0
    for e1 in mc:
        if circuit.is_elem_voltage_defined(e1):
            vde1 += 1
        if isinstance(e1, devices.Capacitor):
            MC[e1.n1 - 1, 0] += 1. if e1.n1 > 0 else 0.
            MC[e1.n2 - 1, 0] -= 1. if e1.n2 > 0 else 0.
        elif isinstance(e1, devices.Inductor):
            MC[nodes_m1 + vde1] += -1.
        elif calc_zeros and e1.part_id == input_source:
            if isinstance(e1, devices.VSource):
                MC[nodes_m1 + vde1] += -1.
            elif isinstance(e1, devices.ISource):
                MC[e1.n1 - 1, 0] += 1. if e1.n1 > 0 else 0.
                MC[e1.n2 - 1, 0] -= 1. if e1.n2 > 0 else 0.
            else:
                raise Exception("Unknown input source type %s" % input_source)
        else:
            continue
        TV = -1. * np.dot(MNAinv, MC)
        dei_victim = 0
        vde2 = -1
        for e2 in mc:
            if circuit.is_elem_voltage_defined(e2):
                vde2 += 1
            if isinstance(e2, devices.Capacitor):
                v = 0
                if e2.n1:
                    v += TV[e2.n1 - 1, 0]
                if e2.n2:
                    v -= TV[e2.n2 - 1, 0]
            elif isinstance(e2, devices.Inductor):
                v = TV[nodes_m1 + vde2, 0]                
            else:
                continue
            if calc_zeros and e1.part_id == input_source:
                RIIN += [v]
            else:
                if not dei_source:
                    TCM = _enlarge_matrix(TCM)
                TCM[dei_victim, dei_source] = v*e1.value
                dei_victim += 1
        if calc_zeros and e1.part_id == input_source:
            ROUTIN = 0
            o1, o2 = output_port
            if o1:
                ROUTIN += TV[o1 - 1, 0]
            if o2:
               ROUTIN -= TV[o2 - 1, 0]
        else:
            dei_source += 1
        # reset, get ready to restart
        MC[:, :] = 0.
    if TCM is not None:
        if np.linalg.det(TCM):
            poles = 1./(2.*np.pi)*(1./np.linalg.eigvals(TCM) + shift)
        else:
            return calculate_singularities(mc, input_source, output_port, MNA=MNA,
                                           outfile=outfile, 
                                           shift=shift+np.abs(1+np.random.uniform())*1e3)
    else:
        poles = []
    if calc_zeros and TCM is not None:
        # re-loop, get the ROUT elements
        vde1 = -1
        MC = np.zeros((MNA.shape[0] - 1, 1))
        ROUT = []
        for e1 in mc:
            if circuit.is_elem_voltage_defined(e1):
                vde1 += 1
            if isinstance(e1, devices.Capacitor):
                MC[e1.n1 - 1, 0] += 1. if e1.n1 > 0 else 0.
                MC[e1.n2 - 1, 0] -= 1. if e1.n2 > 0 else 0.
            elif isinstance(e1, devices.Inductor):
                MC[nodes_m1 + vde1, 0] += -1.
            else:
                continue
            TV = -1.*np.dot(MNAinv, MC)
            v = 0
            o1, o2 = output_port
            if o1:
                v += TV[o1 - 1, 0]
            if o2:
               v -= TV[o2 - 1, 0]
            ROUT += [v*e1.value]
            # reset, get ready to restart
            MC[:, :] = 0.
        # Reshape the matrices and evaluate the zero correction.
        RIIN = np.array(RIIN).reshape((-1, 1))
        RIIN = np.tile(RIIN, (1, RIIN.shape[0]))
        ROUT = np.diag(np.atleast_1d(np.array(ROUT)))
        if ROUT.any():
            try:
                if not np.all(ROUTIN) or not np.isfinite(ROUTIN):
                    # immediate catch
                    raise ValueError("ROUT-IN is either Inf, NaN or null")
                ZTCM = TCM - np.dot(RIIN, ROUT)/ROUTIN
                if not (np.isfinite(ZTCM).all()):
                    raise ValueError("Array contains infs, NaNs or both")
                    # immediate catch
                eigvals = np.linalg.eigvals(ZTCM)
                if not np.all(eigvals) or not np.isfinite(eigvals).all():
                    # immediate catch
                    raise ValueError("ZTCM eigenvalues contain either Inf, NaN or null values")
                zeros = 1./(2.*np.pi)*(1./np.linalg.eigvals(ZTCM) + shift)
            except ValueError:
                return calculate_singularities(mc, input_source, output_port, 
                                           MNA=MNA, outfile=outfile,
                                           shift=shift+np.abs(np.random.uniform()+1)*1e3)
        elif shift < 1e12:
            return calculate_singularities(mc, input_source, output_port, 
                                       MNA=MNA, outfile=outfile,
                                       shift=shift*np.abs(np.random.uniform()+1)*10)
        else:
            zeros = []
    else:
        zeros = []
    poles = np.array([a for a in poles if np.abs(a) < options.pz_max], dtype=np.complex_)
    zeros = np.array([a for a in zeros if np.abs(a) < options.pz_max], dtype=np.complex_)
    poles = np.sort_complex(poles)
    zeros = np.sort_complex(zeros)
    res = results.pz_solution(mc, poles, zeros, outfile)
    return res

def _check_singularities(res, ref, atol=1e-4, rtol=1e-3):
    ref = copy.copy(ref)
    for i in res:
        success = False
        for si, s in enumerate(ref):
            if np.allclose(i, s, atol=atol, rtol=rtol):
                success = True
                break
            assert si != len(ref) - 1
        ref.pop(si)
    assert not len(ref)

